import math
from copy import deepcopy

import plotext_plus._utility as ut
from plotext_plus._build import BuildClass
from plotext_plus._default import DefaultMonitorClass
from plotext_plus._matrix import MatrixClass

# This file defines the monitor class, i.e. the plot, where actual data is plotted; the plot is build separately in the build class for clarity; here only the main tools and drawing methods are written


class MonitorClass(BuildClass):

    def __init__(self):
        self.default = DefaultMonitorClass()  # default values
        self.labels_init()
        self.axes_init()
        self.color_init()
        self.data_init()
        self.matrix = MatrixClass()

    def copy(self):  # to deep copy
        return deepcopy(self)

    def set_size(
        self, size
    ):  # called externally by the figure containing it, to pass the size
        self.size = size

    def set_date(
        self, date
    ):  # called externally by the figure containing it, to pass the date tools so that they share the same settings
        self.date = date

    ##############################################
    #########    Internal Variables    ###########
    ##############################################

    def labels_init(self):
        self.title = None
        self.xlabel = [None, None]
        self.ylabel = [None, None]

    def axes_init(self):
        self.xscale = [self.default.xscale[0]] * 2  # the scale on x axis
        self.yscale = [self.default.xscale[0]] * 2

        self.xticks = self.default.xticks  # xticks coordinates for both axes
        self.xlabels = self.default.xticks[:]  # xlabels for both axes
        self.xfrequency = (
            self.default.xfrequency
        )  # lower and upper xaxes ticks frequency
        self.xdirection = self.default.xdirection

        self.yticks = self.default.yticks
        self.ylabels = self.default.yticks[:]
        self.yfrequency = (
            self.default.yfrequency
        )  # left and right yaxes ticks frequency
        self.ydirection = self.default.ydirection

        self.xaxes = self.default.xaxes  # whatever to show the lower and upper x axis
        self.yaxes = self.default.yaxes  # whatever to show the left and right y axis

        self.grid = (
            self.default.grid
        )  # whatever to show the horizontal and vertical grid lines

    def color_init(self):
        self.set_theme("default")

    def data_init(self):
        self.xlim = [
            [None, None],
            [None, None],
        ]  # the x axis plot limits for lower and upper xside
        self.ylim = [
            [None, None],
            [None, None],
        ]  # the y axis plot limits for left and right yside

        self.fast_plot = False
        self.lines_init()
        self.text_init()
        self.draw_init()

    def lines_init(self):
        self.vcoord = [
            [],
            [],
        ]  # those are user defined extra grid lines, vertical or horizontal, for each axis
        self.hcoord = [[], []]
        self.vcolors = [[], []]  # their color
        self.hcolors = [[], []]

    def text_init(self):
        self.text = []
        self.tx = []
        self.ty = []
        self.txside = []
        self.tyside = []
        self.torien = []
        self.talign = []
        self.tfull = []
        self.tback = []
        self.tstyle = []

    def draw_init(self):  # Variables Set with Draw internal Arguments
        self.xside = (
            []
        )  # which side the x axis should go, for each plot (lower or upper)
        self.yside = (
            []
        )  # which side the y axis should go, for each plot (left or right)

        self.x = []  # list of x coordinates
        self.y = []  # list of y coordinates
        self.x_date = [False, False]  # True if x axis is for date time plots
        self.y_date = [False, False]
        self.signals = 0  # number of signals to plot

        self.lines = []  # whatever to draw lines between points

        self.marker = []  # list of markers used for each plot
        self.color = []  # list of marker colors used for each plot
        self.past_colors = []
        self.style = []

        self.fillx = []  # fill data vertically (till x axis)
        self.filly = []  # fill data horizontally (till y axis)

        self.label = []  # subplot list of labels

    ##############################################
    #######    External Set Functions    #########
    ##############################################

    def set_title(self, title=None):
        self.title = self.set_label(title)

    def set_xlabel(self, label=None, xside=None):
        pos = self.xside_to_pos(xside)
        self.xlabel[pos] = self.set_label(label)

    def set_ylabel(self, label=None, yside=None):
        pos = self.yside_to_pos(yside)
        self.ylabel[pos] = self.set_label(label)

    def set_xlim(self, left=None, right=None, xside=None):
        left = self.date.string_to_time(left) if isinstance(left, str) else left
        right = self.date.string_to_time(right) if isinstance(right, str) else right
        left = None if left is None else float(left)
        right = None if right is None else float(right)
        xlim = [left, right]
        xlim = xlim if None in xlim else [min(xlim), max(xlim)]
        pos = self.xside_to_pos(xside)
        self.xlim[pos] = xlim

    def set_ylim(self, lower=None, upper=None, yside=None):
        lower = self.date.string_to_time(lower) if isinstance(lower, str) else lower
        upper = self.date.string_to_time(upper) if isinstance(upper, str) else upper
        lower = None if lower is None else float(lower)
        upper = None if upper is None else float(upper)
        ylim = [lower, upper]
        ylim = ylim if None in ylim else [min(ylim), max(ylim)]
        pos = self.yside_to_pos(yside)
        self.ylim[pos] = ylim

    def set_xscale(self, scale=None, xside=None):
        default_case = scale is None or scale not in self.default.xscale
        scale = self.default.xscale[0] if default_case else scale
        pos = self.xside_to_pos(xside)
        self.xscale[pos] = scale

    def set_yscale(self, scale=None, yside=None):
        default_case = scale is None or scale not in self.default.yscale
        scale = self.default.yscale[0] if default_case else scale
        pos = self.yside_to_pos(yside)
        self.yscale[pos] = scale

    def set_xticks(self, ticks=None, labels=None, xside=None):
        pos = self.xside_to_pos(xside)
        ticks = self.default.xticks[pos] if ticks is None else list(ticks)
        string_ticks = any(isinstance(el, str) for el in ticks)
        labels = ticks if string_ticks and labels is None else labels
        ticks = self.date.strings_to_time(ticks) if string_ticks else ticks
        labels = ut.get_labels(ticks) if labels is None else list(labels)
        labels = list(map(str, labels))
        ticks, labels = ut.brush(ticks, labels)
        self.xticks[pos] = ticks
        self.xlabels[pos] = labels
        self.xfrequency[pos] = self.xfrequency[pos] if ticks is None else len(ticks)

    def set_yticks(self, ticks=None, labels=None, yside=None):
        pos = self.yside_to_pos(yside)
        ticks = self.default.yticks[pos] if ticks is None else list(ticks)
        string_ticks = any(isinstance(el, str) for el in ticks)
        labels = ticks if string_ticks and labels is None else labels
        ticks = self.date.strings_to_time(ticks) if string_ticks else ticks
        labels = ut.get_labels(ticks) if labels is None else list(labels)
        labels = list(map(str, labels))
        ticks, labels = ut.brush(ticks, labels)
        self.yticks[pos] = ticks
        self.ylabels[pos] = labels
        self.yfrequency[pos] = self.yfrequency[pos] if ticks is None else len(ticks)

    def set_xfrequency(self, frequency=None, xside=None):
        pos = self.xside_to_pos(xside)
        frequency = (
            self.default.xfrequency[pos] if frequency is None else int(frequency)
        )
        self.xfrequency[pos] = frequency

    def set_yfrequency(self, frequency=None, yside=None):
        pos = self.yside_to_pos(yside)
        frequency = (
            self.default.yfrequency[pos] if frequency is None else int(frequency)
        )
        self.yfrequency[pos] = frequency

    def set_xreverse(self, reverse=None, xside=None):
        pos = self.xside_to_pos(xside)
        direction = (
            self.default.xdirection[pos]
            if reverse is None
            else 2 * int(not reverse) - 1
        )
        self.xdirection[pos] = direction

    def set_yreverse(self, reverse=None, yside=None):
        pos = self.yside_to_pos(yside)
        direction = (
            self.default.ydirection[pos]
            if reverse is None
            else 2 * int(not reverse) - 1
        )
        self.ydirection[pos] = direction

    def set_xaxes(self, lower=None, upper=None):
        self.xaxes[0] = self.default.xaxes[0] if lower is None else bool(lower)
        self.xaxes[1] = self.default.xaxes[1] if upper is None else bool(upper)

    def set_yaxes(self, left=None, right=None):
        self.yaxes[0] = self.default.yaxes[0] if left is None else bool(left)
        self.yaxes[1] = self.default.yaxes[1] if right is None else bool(right)

    def set_frame(self, frame=None):
        self.set_xaxes(frame, frame)
        self.set_yaxes(frame, frame)

    def set_grid(self, horizontal=None, vertical=None):
        horizontal = self.default.grid[0] if horizontal is None else bool(horizontal)
        vertical = self.default.grid[1] if vertical is None else bool(vertical)
        self.grid = [horizontal, vertical]

    def set_color(self, color=None):
        color = color if ut.is_color(color) else None
        return self.default.canvas_color if color is None else color

    def set_canvas_color(self, color=None):
        self.canvas_color = self.set_color(color)

    def set_axes_color(self, color=None):
        self.axes_color = self.set_color(color)

    def set_ticks_color(self, color=None):
        self.ticks_color = self.set_color(color)

    def set_ticks_style(self, style=None):
        style = style if ut.is_style(style) else None
        style = self.default.ticks_style if style is None else ut.clean_styles(style)
        self.ticks_style = style

    def set_theme(self, theme=None):
        theme = "default" if theme is None or theme not in ut.themes else theme
        self._set_theme(*ut.themes[theme])

    def clear_color(self):
        self.set_theme("clear")

    ##############################################
    #######    Set Functions Utilities    ########
    ##############################################

    def set_label(self, label=None):
        label = None if label is None else str(label).strip()
        spaces = ut.only_spaces(label)
        label = None if spaces else label
        return label

    def correct_xside(self, xside=None):  # from axis side to position
        xaxis = self.default.xside
        xside = (
            xaxis[xside - 1]
            if isinstance(xside, int) and 1 <= xside <= 2
            else (
                xaxis[0]
                if xside is None or xside.strip() not in xaxis
                else xside.strip()
            )
        )
        return xside

    def correct_yside(self, yside=None):
        yaxis = self.default.yside
        yside = (
            yaxis[yside - 1]
            if isinstance(yside, int) and 1 <= yside <= 2
            else (
                yaxis[0]
                if yside is None or yside.strip() not in yaxis
                else yside.strip()
            )
        )
        return yside

    def xside_to_pos(self, xside=None):  # from axis side to position
        xside = self.correct_xside(xside)
        pos = self.default.xside.index(xside)
        return pos

    def yside_to_pos(self, yside=None):
        yside = self.correct_yside(yside)
        pos = self.default.yside.index(yside)
        return pos

    def _set_theme(
        self, canvas_color, axes_color, ticks_color, ticks_style, color_sequence
    ):
        self.canvas_color = canvas_color
        self.axes_color = axes_color
        self.ticks_color = ticks_color
        self.ticks_style = ticks_style
        self.color_sequence = color_sequence

    ##############################################
    ##########    Draw() Function    #############
    ##############################################

    def draw(
        self, *args, **kwargs
    ):  # from draw() comes directly the functions scatter() and plot()
        self.add_xside(kwargs.get("xside"))
        self.add_yside(kwargs.get("yside"))
        self.add_data(*args)
        self.add_lines(kwargs.get("lines"))
        self.add_markers(kwargs.get("marker"))
        self.add_colors(kwargs.get("color"))
        self.add_styles(kwargs.get("style"))
        self.add_fillx(kwargs.get("fillx"))
        self.add_filly(kwargs.get("filly"))
        self.add_label(kwargs.get("label"))

    ##############################################
    #######    Draw() Called Functions    ########
    ##############################################

    def add_xside(self, xside=None):
        xside = self.correct_xside(xside)
        self.xside.append(xside)

    def add_yside(self, yside=None):
        yside = self.correct_yside(yside)
        self.yside.append(yside)

    def add_data(self, *args):
        x, y = ut.set_data(*args)
        x, x_date = self.to_time(x)
        y, y_date = self.to_time(y)
        self.x_date[self.xside_to_pos(self.xside[-1])] = x_date
        self.y_date[self.yside_to_pos(self.yside[-1])] = y_date
        self.x.append(x)
        self.y.append(y)
        self.signals += 1

    def add_lines(self, lines):
        lines = self.default.lines if lines is None else bool(lines)
        self.lines.append(lines)

    def add_markers(self, marker=None):
        single_marker = isinstance(marker, str) or marker is None
        marker = (
            self.check_marker(marker)
            if single_marker
            else list(map(self.check_marker, marker))
        )
        length = len(self.x[-1])
        marker = ut.to_list(marker, length)
        self.marker.append(marker)

    def add_colors(self, color=None):
        list_color = isinstance(color, list)
        color = (
            list(map(self.check_color, color))
            if list_color
            else self.check_color(color)
        )
        length = len(self.x[-1])
        self.past_colors = (
            self.past_colors + [color]
            if color not in self.past_colors
            else self.past_colors
        )
        color = ut.to_list(color, length)
        self.color.append(color)

    def add_styles(self, style=None):
        single_style = isinstance(style, str) or style is None
        style = (
            self.check_style(style)
            if single_style
            else list(map(self.check_style, style))
        )
        length = len(self.x[-1])
        style = ut.to_list(style, length)
        self.style.append(style)

    def add_fillx(self, fillx=None):
        fillx = self.check_fill(fillx)
        self.fillx.append(fillx)

    def add_filly(self, filly=None):
        filly = self.check_fill(filly)
        self.filly.append(filly)

    def add_label(self, label=None):
        spaces = ut.only_spaces(label)
        label = (
            self.default.label if label is None or spaces else str(label).strip()
        )  # strip to remove spaces before and after
        self.label.append(label)
        # figure.subplot.label_show.append(default.label_show)

    ##############################################
    ######    Draw() Functions Utilities   #######
    ##############################################

    def to_time(self, data):
        dates = any(isinstance(el, str) for el in data)
        data = self.date.strings_to_time(data) if dates else data
        return data, dates

    def check_marker(self, marker=None):
        marker = None if marker is None else str(marker)
        marker = self.default.marker if marker is None else marker
        marker = ut.marker_codes.get(marker, marker)
        marker = marker if marker in ut.hd_symbols else marker[0]
        return marker

    def check_color(self, color=None):
        color = color if ut.is_color(color) else None
        color = self.next_color() if color is None else color
        return color

    def next_color(self):
        color = ut.difference(self.color_sequence, self.past_colors)
        color = color[0] if len(color) > 0 else self.color_sequence[0]
        return color

    def check_style(self, style=None):
        style = None if style is None else str(style)
        style = style if ut.is_style(style) else ut.no_color
        return style

    def check_fill(self, fill=None):
        fill = self.default.fill if fill is None else fill
        fill = (
            False
            if isinstance(fill, str) and fill != self.default.fill_internal
            else fill
        )
        fill = 0 if fill is True else fill
        return fill

    ##############################################
    ######    Other Plotting Functions    ########
    ##############################################

    def draw_bar(
        self,
        *args,
        marker=None,
        color=None,
        fill=None,
        width=None,
        orientation=None,
        minimum=None,
        offset=None,
        reset_ticks=None,
        xside=None,
        yside=None,
        label=None,
    ):
        x, y = ut.set_data(*args)
        marker = self.default.bar_marker if marker is None else marker
        fill = self.default.bar_fill if fill is None else fill
        width = self.default.bar_width if width is None else width
        width = 1 if width > 1 else 0 if width < 0 else width
        orientation = self.check_orientation(orientation, 1)
        minimum = 0 if minimum is None else minimum
        offset = 0 if offset is None else offset
        reset_ticks = True if reset_ticks is None else reset_ticks

        x_string = any(isinstance(el, str) for el in x)  # if x are strings
        x_length = len(x)
        xticks = range(1, x_length + 1) if x_string else x
        xlabels = x if x_string else map(str, x)
        x = xticks if x_string else x
        x = [el + offset for el in x]
        xbar, ybar = ut.bars(x, y, width, minimum)
        xbar, ybar = [xbar, ybar] if orientation[0] == "v" else [ybar, xbar]
        (
            (
                self.set_xticks(xticks, xlabels, xside)
                if orientation[0] == "v"
                else self.set_yticks(xticks, xlabels, yside)
            )
            if reset_ticks
            else None
        )

        firstbar = min(
            [b for b in range(len(x)) if ybar[b][1] != 0], default=0
        )  # finds the position of the first non zero bar

        for b in range(len(x)):
            xb = xbar[b]
            yb = ybar[b]
            plot_label = label if b == firstbar else None
            plot_color = color if b == 0 else self.color[-1]
            nobar = (yb[1] == 0 and orientation[0] == "v") or (
                xb[1] == 0 and orientation[0] == "h"
            )
            plot_marker = " " if nobar else marker
            plot_color = color if b == 0 else self.color[-1][-1]
            self.draw_rectangle(
                xb,
                yb,
                xside=xside,
                yside=yside,
                lines=True,
                marker=plot_marker,
                color=plot_color,
                fill=fill,
                label=plot_label,
            )

    def draw_multiple_bar(
        self,
        *args,
        marker=None,
        color=None,
        fill=None,
        width=None,
        orientation=None,
        minimum=None,
        offset=None,
        reset_ticks=None,
        xside=None,
        yside=None,
        labels=None,
    ):
        x, y_values = ut.set_multiple_bar_data(*args)
        ly = len(y_values)
        width = self.default.bar_width if width is None else width
        marker = [marker] * ly if marker is None or not isinstance(marker, list) else marker
        color = [color] * ly if color is None else color
        labels = [labels] * ly if labels is None else labels
        width = width / ly if ly != 0 else 0
        offset = (
            ut.linspace(-1 / 2 + 1 / (2 * ly), 1 / 2 - 1 / (2 * ly), ly)
            if ly != 0
            else []
        )

        for i in range(ly):
            self.draw_bar(
                x,
                y_values[i],
                marker=marker[i],
                color=color[i],
                fill=fill,
                width=width,
                orientation=orientation,
                minimum=minimum,
                offset=offset[i],
                xside=xside,
                yside=yside,
                label=labels[i],
                reset_ticks=reset_ticks,
            )

    def draw_stacked_bar(
        self,
        *args,
        marker=None,
        color=None,
        fill=None,
        width=None,
        orientation=None,
        minimum=None,
        offset=None,
        reset_ticks=None,
        xside=None,
        yside=None,
        labels=None,
    ):
        x, y_values = ut.set_multiple_bar_data(*args)
        ly = len(y_values)
        marker = [marker] * ly if marker is None or not isinstance(marker, list) else marker
        color = [color] * ly if color is None else color
        labels = [labels] * ly if labels is None else labels
        y_values = ut.transpose([ut.cumsum(el) for el in ut.transpose(y_values)])
        for i in range(ly - 1, -1, -1):
            self.draw_bar(
                x,
                y_values[i],
                xside=xside,
                yside=yside,
                marker=marker[i],
                color=color[i],
                fill=fill,
                width=width,
                orientation=orientation,
                label=labels[i],
                minimum=minimum,
                reset_ticks=reset_ticks,
            )

    def draw_hist(
        self,
        data,
        bins=None,
        marker=None,
        color=None,
        fill=None,
        norm=None,
        width=None,
        orientation=None,
        minimum=None,
        xside=None,
        yside=None,
        label=None,
    ):
        bins = self.default.hist_bins if bins is None else bins
        norm = False if norm is None else norm
        x, y = ut.hist_data(data, bins, norm)
        self.draw_bar(
            x,
            y,
            xside=xside,
            yside=yside,
            marker=marker,
            color=color,
            fill=fill,
            width=width,
            orientation=orientation,
            label=label,
            minimum=None,
            reset_ticks=False,
        )

    def draw_candlestick(
        self,
        dates,
        data,
        colors=None,
        orientation=None,
        xside=None,
        yside=None,
        label=None,
    ):
        orientation = self.check_orientation(orientation, 1)
        markers = ["sd", "│", "─"]  # if markers is None else markers
        colors = ["green", "red"] if colors is None else colors
        color = []
        ln = len(dates)
        data = (
            {"Open": [], "Close": [], "High": [], "Low": []} if len(data) == 0 else data
        )
        open_vals = data["Open"]
        close_vals = data["Close"]
        high_vals = data["High"]
        low_vals = data["Low"]
        for i in range(ln):
            d = dates[i]
            o, c, h, low_val = open_vals[i], close_vals[i], high_vals[i], low_vals[i]
            color = colors[0] if c > o else colors[1]
            m, max_val = min(o, c), max(o, c)
            lab = label if i == 0 else None
            if orientation in ["v", "vertical"]:
                self.draw(
                    [d, d],
                    [max_val, h],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[1],
                    lines=True,
                )
                self.draw(
                    [d, d],
                    [low_val, m],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[1],
                    lines=True,
                )
                self.draw(
                    [d, d],
                    [m, max_val],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[0],
                    lines=True,
                    label=lab,
                )
            elif orientation in ["h", "horizontal"]:
                self.draw(
                    [max_val, h],
                    [d, d],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[2],
                    lines=True,
                )
                self.draw(
                    [low_val, m],
                    [d, d],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[2],
                    lines=True,
                )
                self.draw(
                    [m, max_val],
                    [d, d],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[0],
                    lines=True,
                    label=lab,
                )

    def draw_box(
        self,
        *args,
        xside=None,
        yside=None,
        orientation=None,
        colors=None,
        label=None,
        fill=None,
        width=None,
        minimum=None,
        offset=None,
        reset_ticks=None,
        quintuples=None,
    ):
        x, y = ut.set_data(*args)
        fill = self.default.bar_fill if fill is None else fill
        width = self.default.bar_width if width is None else width
        width = 1 if width > 1 else 0 if width < 0 else width
        orientation = self.check_orientation(orientation, 1)
        minimum = 0 if minimum is None else minimum
        offset = 0 if offset is None else offset
        reset_ticks = True if reset_ticks is None else reset_ticks
        colors = ["green", "red"] if colors is None else colors
        quintuples = False if quintuples is None else quintuples

        x_string = any(isinstance(el, str) for el in x)  # if x are strings
        x_length = len(x)
        xticks = range(1, x_length + 1) if x_string else x
        xlabels = x if x_string else map(str, x)
        x = xticks if x_string else x
        x = [el + offset for el in x]
        (
            (
                self.set_xticks(xticks, xlabels, xside)
                if orientation[0] == "v"
                else self.set_yticks(xticks, xlabels, yside)
            )
            if reset_ticks
            else None
        )
        if quintuples:
            # todo: check y is aligned.
            _, _, _, _, _, c, xbar = ut.box(x, y, width, minimum)
            q1, q2, q3, max_, min_ = [], [], [], [], []
            for d in y:
                max_.append(d[0])
                q3.append(d[1])
                q2.append(d[2])
                q1.append(d[3])
                min_.append(d[4])
        else:
            q1, q2, q3, max_, min_, c, xbar = ut.box(x, y, width, minimum)
        markers = ["sd", "│", "─"]  # if markers is None else markers

        for i in range(x_length):
            lab = label if i == 0 else None
            color = colors[0]
            mcolor = colors[1]
            d, low_val, h, m, median, max_val = c[i], min_[i], max_[i], q1[i], q2[i], q3[i]
            (max_val - m) / 30
            if orientation in ["v", "vertical"]:
                self.draw(
                    [d, d],
                    [max_val, h],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[1],
                    lines=True,
                )
                self.draw(
                    [d, d],
                    [low_val, m],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[1],
                    lines=True,
                )
                self.draw_rectangle(
                    xbar[i],
                    [m, max_val],
                    xside=xside,
                    yside=yside,
                    lines=True,
                    color=color,
                    fill=fill,
                    marker=markers[0],
                    label=lab,
                )
                self.draw_rectangle(
                    xbar[i],
                    [median, median],
                    xside=xside,
                    yside=yside,
                    lines=True,
                    color=mcolor,
                    fill=fill,
                    marker=markers[2],
                )
                # self.draw([d, d], [m, M], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)
                # self.draw(xbar[i], [E, E], xside = xside, yside = yside, color = mcolor, marker = markers[0], lines = False)
            elif orientation in ["h", "horizontal"]:
                self.draw(
                    [max_val, h],
                    [d, d],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[2],
                    lines=True,
                )
                self.draw(
                    [low_val, m],
                    [d, d],
                    xside=xside,
                    yside=yside,
                    color=color,
                    marker=markers[2],
                    lines=True,
                )
                self.draw_rectangle(
                    [m, max_val],
                    xbar[i],
                    xside=xside,
                    yside=yside,
                    lines=True,
                    color=color,
                    fill=fill,
                    marker=markers[0],
                    label=lab,
                )
                self.draw_rectangle(
                    [median, median],
                    xbar[i],
                    xside=xside,
                    yside=yside,
                    lines=True,
                    color=mcolor,
                    fill=fill,
                    marker=markers[1],
                )
                # self.draw([m, M], [d, d], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)
                # self.draw([E, E], [d, d], xside = xside, yside = yside, color = 'red', marker = markers[0], lines = True)

    ##############################################
    ###########    Plotting Tools    #############
    ##############################################

    def draw_error(
        self,
        *args,
        xerr=None,
        yerr=None,
        color=None,
        xside=None,
        yside=None,
        label=None,
    ):
        x, y = ut.set_data(*args)
        x_length = len(x)
        xerr = [0] * x_length if xerr is None else xerr
        yerr = [0] * x_length if yerr is None else yerr
        for i in range(x_length):
            col = self.color[-1][-1] if i > 0 else color
            self.draw(
                [x[i], x[i]],
                [y[i] - yerr[i] / 2, y[i] + yerr[i] / 2],
                xside=xside,
                yside=yside,
                marker="│",
                color=col,
                lines=True,
            )
            col = self.color[-1][-1] if i == 0 else col
            self.draw(
                [x[i] - xerr[i] / 2, x[i] + xerr[i] / 2],
                [y[i], y[i]],
                xside=xside,
                yside=yside,
                marker="─",
                color=col,
                lines=True,
            )
            self.draw(
                [x[i]],
                [y[i]],
                xside=xside,
                yside=yside,
                marker="┼",
                color=col,
                lines=True,
            )

    def draw_event_plot(
        self, data, marker=None, color=None, orientation=None, side=None
    ):
        x, y = data, [1.1] * len(data)
        orientation = self.check_orientation(orientation, 1)
        if orientation in ["v", "vertical"]:
            self.draw(x, y, xside=side, marker=marker, color=color, fillx=True)
            self.set_ylim(0, 1)
            self.set_yfrequency(0)
        else:
            self.draw(y, x, yside=side, marker=marker, color=color, filly=True)
            self.set_xlim(0, 1)
            self.set_xfrequency(0)

    def draw_vertical_line(self, coordinate, color=None, xside=None):
        coordinate = (
            self.date.string_to_time(coordinate)
            if isinstance(coordinate, str)
            else coordinate
        )
        pos = self.xside_to_pos(xside)
        self.vcoord[pos].append(coordinate)
        color = self.ticks_color if color is None else color
        self.vcolors[pos].append(self.check_color(color))

    def draw_horizontal_line(self, coordinate, color=None, yside=None):
        coordinate = (
            self.date.string_to_time(coordinate)
            if isinstance(coordinate, str)
            else coordinate
        )
        pos = self.xside_to_pos(yside)
        self.hcoord[pos].append(coordinate)
        color = self.ticks_color if color is None else color
        self.hcolors[pos].append(self.check_color(color))

    def draw_text(
        self,
        text,
        x,
        y,
        xside=None,
        yside=None,
        color=None,
        background=None,
        style=None,
        orientation=None,
        alignment=None,
    ):
        orientation = self.check_orientation(orientation)
        text = text if orientation is self.default.orientation[0] else text[::-1]
        self.text.append(str(text))
        x = self.date.string_to_time(x) if isinstance(x, str) else x
        y = self.date.string_to_time(y) if isinstance(y, str) else y
        self.tx.append(x)
        self.ty.append(y)
        self.txside.append(self.correct_xside(xside))
        self.tyside.append(self.correct_yside(yside))
        color = self.next_color() if color is None or not ut.is_color(color) else color
        background = (
            self.canvas_color
            if background is None or not ut.is_color(background)
            else background
        )
        self.tfull.append(color)
        self.tback.append(background)
        self.tstyle.append(self.check_style(style))
        alignment = self.check_alignment(alignment)
        self.torien.append(orientation)
        self.talign.append(alignment)

    def draw_rectangle(
        self,
        x=None,
        y=None,
        marker=None,
        color=None,
        lines=None,
        fill=None,
        reset_lim=False,
        xside=None,
        yside=None,
        label=None,
    ):
        x = [0, 1] if x is None or len(x) < 2 else x
        y = [0, 1] if y is None or len(y) < 2 else y
        xpos = self.xside_to_pos(xside)
        self.yside_to_pos(yside)
        lines = True if lines is None else lines
        fill = False if fill is None else fill
        xm = min(x)
        x_max = max(x)
        ym = min(y)
        y_max = max(y)
        dx = abs(x_max - xm)
        dy = abs(y_max - ym)
        if reset_lim:
            self.xlim[xpos] = [xm - 0.5 * dx, x_max + 0.5 * dx]
            self.ylim[xpos] = [ym - 0.5 * dy, y_max + 0.5 * dy]
        x, y = [xm, xm, x_max, x_max, xm], [ym, y_max, y_max, ym, ym]
        self.draw(
            x,
            y,
            xside=xside,
            yside=yside,
            lines=True if fill else lines,
            marker=marker,
            color=color,
            fillx="internal" if fill else False,
            filly=False,
            label=label,
        )

    def draw_polygon(
        self,
        x=None,
        y=None,
        radius=None,
        sides=None,
        marker=None,
        color=None,
        lines=None,
        fill=None,
        reset_lim=False,
        xside=None,
        yside=None,
        label=None,
    ):
        x = 0 if x is None else x
        y = 0 if y is None else y
        radius = 1 if radius is None else abs(int(radius))
        sides = 3 if sides is None else max(3, int(abs(sides)))
        xpos = self.xside_to_pos(xside)
        self.yside_to_pos(yside)
        lines = True if lines is None else lines
        fill = False if fill is None else fill

        alpha = 2 * math.pi / sides
        init = (
            alpha / 2 + math.pi / 2
            if sides % 2 == 0
            else alpha / 4 * ((-1) ** (sides // 2))
        )  # * math.pi #- ((-1) ** (sides)) * alpha / 4
        # init = 0 * init
        def get_point(i):
            return [
                    x + math.cos(alpha * i + init) * radius,
                    y + math.sin(alpha * i + init) * radius,
                ]
        # the rounding is needed so that results like 9.9999 are rounded to 10 and display as same coordinate in the plot, otherwise the floor function will turn 9.999 into 9
        points = [get_point(i) for i in range(sides + 1)]
        if reset_lim:
            self.xlim[xpos] = [x - 1.5 * radius, x + 1.5 * radius]
            self.ylim[xpos] = [y - 1.5 * radius, y + 1.5 * radius]
        self.draw(
            *ut.transpose(points),
            xside=xside,
            yside=yside,
            lines=True if fill else lines,
            marker=marker,
            color=color,
            fillx="internal" if fill else False,
            filly=False,
            label=label,
        )

    def draw_confusion_matrix(
        self, actual, predicted, color=None, style=None, labels=None
    ):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)

        length = len(actual)
        n_labels = sorted(ut.no_duplicates(actual))
        labels = n_labels if labels is None else list(labels)
        label_count = len(n_labels)

        def get_sum(a, p):
            return sum(
                    [actual[i] == a and predicted[i] == p for i in range(length)]
                )
        cmatrix = [
            [get_sum(n_labels[r], n_labels[c]) for c in range(label_count)] for r in range(label_count)
        ]
        cm = ut.join(cmatrix)
        m, max_val, t = min(cm), max(cm), sum(cm)

        lm = 253
        light_max = 80
        def to_255(light_val):
            return round(
                    lm + (light_max - lm) * (light_val - m) / (max_val - m)
                )  # light_val=m -> lm; light_val=max_val->light_max
        def to_color(light_val):
            return tuple([to_255(light_val)] * 3)
        def to_text(n):
            return str(round(n, 2)) + " - " + str(round(100 * n / t, 2)) + "%"
        for r in range(label_count):
            for c in range(label_count):
                count = cmatrix[r][c]
                col = to_color(count)
                self.draw_rectangle(
                    [c - 0.5, c + 0.5], [r - 0.5, r + 0.5], color=col, fill=True
                )
                self.draw_text(
                    to_text(count), c, r, color=color, background=col, style=style
                )

        self.set_yreverse(True)
        self.set_xticks(n_labels, labels)
        self.set_yticks(n_labels, labels)
        self.set_ticks_color(color)
        self.set_ticks_style(style)
        self.set_axes_color("default")
        self.set_canvas_color("default")
        self.set_title("Confusion Matrix")
        self.set_xlabel("Predicted")
        self.set_ylabel("Actual")

    def draw_indicator(self, value, label=None, color=None, style=None):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)

        self.set_title(label)
        self.set_ticks_color(color)
        self.set_ticks_style(style)
        self.set_axes_color("default")
        self.set_canvas_color("default")
        self.set_xfrequency(0)
        self.set_yfrequency(0)

        self.draw_text(str(value), 0, 0, color=color, style=style, alignment="center")

    ##############################################
    ##############    2D Plots    ################
    ##############################################

    def draw_matrix(self, matrix, marker=None, style=None, fast=False):
        matrix = [row.copy() for row in matrix]
        marker = [marker] if not isinstance(marker, list) else marker
        marker = [
            (
                self.check_marker("sd")
                if el in ut.join([None, ut.hd_symbols])
                else self.check_marker(el)
            )
            for el in marker
        ]
        style = ut.no_color if style is None else self.check_style(style)
        cols, rows = ut.matrix_size(matrix)
        rows = 0 if cols == 0 else rows
        matrix = (
            matrix
            if rows * cols != 0 and ut.is_rgb_color(matrix[0][0])
            else ut.turn_gray(matrix)
        )
        marker = ut.repeat(marker, cols)
        if not fast:
            for r in range(rows):
                xyc = [(c, r, matrix[rows - 1 - r][c]) for c in range(cols)]
                x, y, color = ut.transpose(xyc, 3)
                self.draw(x, y, marker=marker, color=color, style=style)
            self.set_canvas_color("black")
            self.set_xlabel("column")
            self.set_ylabel("row")
            xf, yf = min(self.xfrequency[0], cols), min(self.yfrequency[0], rows)
            xt = ut.linspace(0, cols - 1, xf)
            xl = ut.get_labels([el + 1 for el in xt])
            yt = ut.linspace(0, rows - 1, yf)
            yl = ut.get_labels([rows - el for el in yt])
            self.set_xticks(xt, xl)
            self.set_yticks(yt, yl)
        else:  # if fast
            for r in range(rows):
                for c in range(cols):
                    ansi = ut.colors_to_ansi(matrix[r][c], style, "black")
                    matrix[r][c] = ansi + marker[c] + ut.ansi_end
            self.matrix.canvas = "\n".join(["".join(row) for row in matrix])
            self.fast_plot = True

    def draw_pie(
        self,
        labels,
        values,
        colors=None,
        radius=None,
        show_values=True,
        show_percentages=True,
        title=None,
        show_values_on_slices=False,
        donut=False,
        remaining_color=None,
    ):
        """
        Draw a pie chart using filled colored segments and a legend.

        Args:
            donut (bool): If True, creates a doughnut chart with hollow center (inner radius = 1/3 outer radius)
            remaining_color (str): If specified, colors the remaining slice with this color instead of leaving it as spaces
        """
        import math

        # Input validation
        if len(labels) != len(values):
            raise ValueError("Labels and values must have the same length")

        # Calculate total and percentages
        total = sum(values)
        if total == 0:
            raise ValueError("Total of values cannot be zero")

        percentages = [(value / total) * 100 for value in values]

        # Default colors if not provided
        if colors is None:
            color_cycle = ["red", "blue", "green", "orange", "magenta", "cyan", "white"]
            colors = [color_cycle[i % len(color_cycle)] for i in range(len(labels))]

        # Default radius - calculate based on available plot space
        if radius is None:
            # Get the actual plot area dimensions
            plot_width, plot_height = self.size

            # Set radius to half of the smaller dimension minus 4 for border margin
            radius = (min(plot_width, plot_height) - 4) / 2.0
            radius = max(radius, 3)  # Ensure minimum radius of 3

        # Center the pie chart
        center_x = 0
        center_y = 0

        # Terminal characters have an aspect ratio of approximately 1.5:1 (height:width)
        # To make circles appear circular, we need to adjust the x-axis scaling
        aspect_ratio = 1.5

        # Remove axes - pie charts don't have them
        self.set_xfrequency(0)
        self.set_yfrequency(0)
        self.set_axes_color("default")
        self.set_canvas_color("default")

        # Collect all points for each segment, then draw each segment in one call
        # Use efficient scanning - just slightly beyond the actual pie radius
        # For doughnuts, use denser scanning to ensure solid ring
        if donut:
            scan_radius_x = int(radius * aspect_ratio * 1.5 + 3)
            scan_radius_y = int(radius * 1.5 + 3)
        else:
            scan_radius_x = int(radius * aspect_ratio * 1.2 + 2)
            scan_radius_y = int(radius * 1.2 + 2)

        # Pre-calculate cumulative angles for segment boundaries
        segment_boundaries = []
        current_cumulative = 0
        for value in values:
            slice_angle = (value / total) * 2 * math.pi
            segment_boundaries.append(
                (current_cumulative, current_cumulative + slice_angle)
            )
            current_cumulative += slice_angle

        # Collect all points for each segment using sets to avoid duplicates
        segment_points = [set() for _ in range(len(labels))]  # One set per segment

        # Use FLOOD FILL approach - systematically fill every position in concentric circles
        # This ensures no gaps by filling from center outward
        for y_offset in range(-scan_radius_y, scan_radius_y + 1):
            for x_offset in range(-scan_radius_x, scan_radius_x + 1):
                # Calculate distance from center with aspect ratio correction
                # Since terminal chars are ~1.5x taller than wide, compress x coordinate
                adjusted_x = x_offset / aspect_ratio
                distance = math.sqrt(adjusted_x * adjusted_x + y_offset * y_offset)

                # For doughnut inner boundary, use elliptical check to create circular appearance
                # The inner boundary should be elliptical in terminal coordinates to appear circular
                inner_radius = radius / 3.0 if donut else 0

                # Check if point is outside the inner ellipse (for circular appearance)
                if donut:
                    # Create elliptical inner boundary: x^2/a^2 + y^2/b^2 > r^2
                    # where a = inner_radius * aspect_ratio, b = inner_radius
                    ellipse_x_term = (x_offset * x_offset) / (
                        inner_radius * aspect_ratio * inner_radius * aspect_ratio
                    )
                    ellipse_y_term = (y_offset * y_offset) / (
                        inner_radius * inner_radius
                    )
                    ellipse_value = ellipse_x_term + ellipse_y_term
                    outside_inner = ellipse_value > 1.0

                else:
                    outside_inner = True

                # Use exact radius to stay within plot boundaries
                threshold = radius

                if distance <= threshold and outside_inner:
                    # Calculate angle for this position using adjusted coordinates
                    angle = math.atan2(y_offset, adjusted_x)
                    if angle < 0:
                        angle += 2 * math.pi

                    # Find which segment this position belongs to using robust angle detection
                    segment_idx = 0
                    found_segment = False
                    epsilon = 0.02  # Even larger epsilon for maximum boundary coverage

                    for i, (start_angle, end_angle) in enumerate(segment_boundaries):
                        # Handle wraparound case for segments that cross 0 degrees
                        if end_angle > 2 * math.pi:
                            wrap_end = end_angle - 2 * math.pi
                            if (
                                angle >= start_angle - epsilon
                                or angle <= wrap_end + epsilon
                            ):
                                segment_idx = i
                                found_segment = True
                                break
                        else:
                            # Use very generous boundary detection
                            # For the last segment, use <= to include the boundary
                            if i == len(segment_boundaries) - 1:
                                if (
                                    start_angle - epsilon
                                    <= angle
                                    <= end_angle + epsilon
                                ):
                                    segment_idx = i
                                    found_segment = True
                                    break
                            else:
                                if start_angle - epsilon <= angle < end_angle + epsilon:
                                    segment_idx = i
                                    found_segment = True
                                    break

                    # If no segment found (due to floating point precision), assign based on closest angle
                    if not found_segment:
                        # Find the segment with the smallest angle distance
                        min_distance = float("inf")
                        for i, (start_angle, end_angle) in enumerate(
                            segment_boundaries
                        ):
                            mid_angle = (start_angle + end_angle) / 2
                            # Handle wraparound for mid angle calculation
                            if end_angle > 2 * math.pi:
                                mid_angle = start_angle + (
                                    (end_angle - start_angle) / 2
                                )
                                if mid_angle > 2 * math.pi:
                                    mid_angle -= 2 * math.pi

                            # Calculate angular distance (accounting for circular nature)
                            angle_diff = abs(angle - mid_angle)
                            if angle_diff > math.pi:
                                angle_diff = 2 * math.pi - angle_diff

                            if angle_diff < min_distance:
                                min_distance = angle_diff
                                segment_idx = i

                    # Add this exact character position to the appropriate segment
                    char_x = center_x + x_offset
                    char_y = center_y + y_offset
                    segment_points[segment_idx].add((char_x, char_y))

        # SECOND PASS: Fill any potential gaps by adding adjacent positions
        # This ensures complete coverage by adding neighboring positions to existing points
        additional_points = [set() for _ in range(len(labels))]
        for segment_idx, points in enumerate(segment_points):
            for x, y in points:
                # Add neighboring positions to ensure no gaps
                for dx in [-1, 0, 1]:
                    for dy in [-1, 0, 1]:
                        neighbor_x = x + dx
                        neighbor_y = y + dy

                        # Check if this neighbor is within the circular area using same logic as main pass
                        x_offset = neighbor_x - center_x
                        y_offset = neighbor_y - center_y
                        adjusted_x = x_offset / aspect_ratio
                        neighbor_distance = math.sqrt(
                            adjusted_x * adjusted_x + y_offset * y_offset
                        )

                        # Use same boundary checks as main algorithm
                        inner_radius = radius / 3.0 if donut else 0

                        # Check if point is outside the inner ellipse (for circular appearance)
                        if donut:
                            # Use same elliptical inner boundary as main pass
                            ellipse_x_term = (x_offset * x_offset) / (
                                inner_radius
                                * aspect_ratio
                                * inner_radius
                                * aspect_ratio
                            )
                            ellipse_y_term = (y_offset * y_offset) / (
                                inner_radius * inner_radius
                            )
                            outside_inner = ellipse_x_term + ellipse_y_term > 1.0
                        else:
                            outside_inner = True

                        # Use exact radius to stay within plot boundaries (same as main pass)
                        threshold = radius

                        if neighbor_distance <= threshold and outside_inner:
                            additional_points[segment_idx].add((neighbor_x, neighbor_y))

        # Merge additional points with main points
        for segment_idx in range(len(labels)):
            segment_points[segment_idx].update(additional_points[segment_idx])

        # Draw each segment using a different approach - draw filled shapes row by row
        # This ensures complete filling without gaps
        for _segment_idx, (points, color) in enumerate(
            zip(segment_points, colors, strict=False)
        ):
            if points:  # Only draw if segment has points
                # Handle remaining_color for single-value pie charts
                if color == "default":
                    if remaining_color is not None:
                        # Use the specified remaining_color instead of default
                        color = remaining_color
                    else:
                        # Skip drawing - leave as spaces (current behavior)
                        continue

                points_list = list(points)

                # Group points by y-coordinate to draw horizontal filled lines
                y_groups = {}
                for x, y in points_list:
                    if y not in y_groups:
                        y_groups[y] = []
                    y_groups[y].append(x)

                # For doughnut charts, use smart filling that avoids the hollow center
                # For regular pie charts, use full horizontal line filling
                if donut:
                    # Smart filling for doughnut charts: fill gaps within ring segments but avoid center
                    for y_coord, x_coords in y_groups.items():
                        if x_coords:
                            x_coords.sort()  # Sort x coordinates

                            # Find continuous segments, avoiding the center gap
                            fill_x_coords = []
                            x_step = 0.5

                            # Determine if this y_coord passes through the hollow center
                            y_offset = y_coord - center_y
                            center_x_range = []

                            # Calculate the x-range that should be hollow at this y-coordinate
                            if abs(y_offset) < inner_radius:
                                # This y-line passes through the hollow center
                                # Calculate x-bounds of the elliptical hollow area
                                ellipse_y_term = (y_offset * y_offset) / (
                                    inner_radius * inner_radius
                                )
                                if ellipse_y_term < 1.0:
                                    ellipse_x_term_needed = 1.0 - ellipse_y_term
                                    max_x_offset = (
                                        math.sqrt(ellipse_x_term_needed)
                                        * inner_radius
                                        * aspect_ratio
                                    )
                                    center_x_min = center_x - max_x_offset
                                    center_x_max = center_x + max_x_offset
                                    center_x_range = [center_x_min, center_x_max]

                            # Fill between consecutive x-coordinates, but avoid the center region
                            i = 0
                            while i < len(x_coords):
                                segment_start = x_coords[i]

                                # Find the end of this continuous segment
                                j = i
                                while j < len(x_coords) - 1:
                                    gap = x_coords[j + 1] - x_coords[j]
                                    # If there's a large gap, this segment ends
                                    if (
                                        gap > 2.0
                                    ):  # Allow small gaps but break on large ones
                                        break
                                    j += 1

                                segment_end = x_coords[j]

                                # Fill this segment, but avoid the center region
                                if center_x_range:
                                    # Split segment around the hollow center
                                    center_min, center_max = center_x_range

                                    # Fill left part (before center)
                                    if segment_start < center_min:
                                        left_end = min(segment_end, center_min)
                                        current_x = segment_start
                                        while current_x <= left_end:
                                            fill_x_coords.append(current_x)
                                            current_x += x_step

                                    # Fill right part (after center)
                                    if segment_end > center_max:
                                        right_start = max(segment_start, center_max)
                                        current_x = right_start
                                        while current_x <= segment_end:
                                            fill_x_coords.append(current_x)
                                            current_x += x_step
                                else:
                                    # No center interference, fill entire segment
                                    current_x = segment_start
                                    while current_x <= segment_end:
                                        fill_x_coords.append(current_x)
                                        current_x += x_step

                                i = j + 1

                            # Draw the filled segments
                            if fill_x_coords:
                                fill_y_coords = [y_coord] * len(fill_x_coords)
                                self.draw(
                                    fill_x_coords,
                                    fill_y_coords,
                                    marker="sd",
                                    color=color,
                                )
                else:
                    # For regular pie charts, use full horizontal line filling
                    for y_coord, x_coords in y_groups.items():
                        if x_coords:
                            x_coords.sort()  # Sort x coordinates
                            x_min, x_max = min(x_coords), max(x_coords)

                            # Create a continuous range of x coordinates to fill the gap
                            if x_max > x_min:
                                # Draw filled horizontal line from x_min to x_max
                                fill_x_coords = []
                                x_step = 0.5  # Smaller step for better coverage
                                current_x = x_min
                                while current_x <= x_max:
                                    fill_x_coords.append(current_x)
                                    current_x += x_step
                                fill_y_coords = [y_coord] * len(fill_x_coords)
                                self.draw(
                                    fill_x_coords,
                                    fill_y_coords,
                                    marker="sd",
                                    color=color,
                                )
                            else:
                                # Single point
                                self.draw([x_min], [y_coord], marker="sd", color=color)

        # Reset cumulative_angle for label drawing
        cumulative_angle = 0
        for _i, (_label, value, percentage, _color) in enumerate(
            zip(labels, values, percentages, colors, strict=False)
        ):
            slice_angle = (value / total) * 2 * math.pi

            # Add value labels on the pie slice (only if show_values_on_slices is True)
            if show_values_on_slices and (show_values or show_percentages):
                # Calculate middle angle of the slice for label placement
                middle_angle = cumulative_angle + slice_angle / 2
                # Position label at 70% of radius for better visibility
                label_radius = radius * 0.7
                label_x = (
                    center_x + (label_radius * math.cos(middle_angle)) * aspect_ratio
                )
                label_y = center_y + label_radius * math.sin(middle_angle)

                # Build label text for the slice
                slice_label = ""
                if show_values and show_percentages:
                    slice_label = f"{value}\n({percentage:.1f}%)"
                elif show_values:
                    slice_label = str(value)
                elif show_percentages:
                    slice_label = f"{percentage:.1f}%"

                # Draw the label on the slice
                self.draw_text(
                    slice_label, label_x, label_y, color="white", alignment="center"
                )

            cumulative_angle += slice_angle

        # Extend the plot area to accommodate legend (calculate before filtering)
        max_text_length = (
            max(
                len(f"{label}: {value} ({percentage:.1f}%)")
                for label, value, percentage in zip(
                    labels, values, percentages, strict=False
                )
                if label.lower() != "remaining"
            )
            if any(label.lower() != "remaining" for label in labels)
            else 20
        )

        # Set plot limits to include legend area (adjust x for aspect ratio)
        x_radius = radius * aspect_ratio
        self.set_xlim(-x_radius - 1, x_radius + max_text_length + 2)
        self.set_ylim(-radius - 1, radius + 1)

        # Create legend positioned in the bottom right corner of the chart
        legend_start_x = x_radius + 1.5
        legend_start_y = -radius + len(labels) * 1.0 - 0.5

        # Filter out "Remaining" labels and default colors for single-value pie charts
        legend_items = []
        for _i, (label, value, percentage, color) in enumerate(
            zip(labels, values, percentages, colors, strict=False)
        ):
            # Handle remaining_color logic for legend
            if color == "default":
                if remaining_color is not None:
                    # Show "Remaining" in legend when remaining_color is specified
                    legend_items.append((label, value, percentage, remaining_color))
                # Skip if no remaining_color (leave as spaces)
            else:
                # Always show non-default colors
                legend_items.append((label, value, percentage, color))

        # Adjust legend positioning for filtered items
        legend_start_y = -radius + len(legend_items) * 1.0 - 0.5

        for i, (label, value, percentage, color) in enumerate(legend_items):
            legend_x = legend_start_x
            legend_y = legend_start_y - i * 1.2  # Space between legend items

            # Draw colored square for legend matching pie chart blocks
            self.draw([legend_x], [legend_y], marker="sd", color=color)

            # Build legend text with colored block prefix
            block_char = "█"  # Solid block character
            legend_text = f"{block_char} {label}"
            if show_values and show_percentages:
                legend_text += f": {value} ({percentage:.1f}%)"
            elif show_values:
                legend_text += f": {value}"
            elif show_percentages:
                legend_text += f": {percentage:.1f}%"

            # Use draw_text for the legend with the same color as the segment
            self.draw_text(legend_text, legend_x, legend_y, color=color)

        # Set title if provided
        if title:
            self.set_title(title)

        # Remove axis labels since pie charts don't need them
        self.set_xlabel("")
        self.set_ylabel("")

    def draw_heatmap(self, dataframe, color=None, style=None):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)

        xlabels = dataframe.columns.tolist()
        ylabels = dataframe.index.tolist()

        cmatrix = dataframe.values.tolist()
        cm = ut.join(cmatrix)
        m, max_val, _t = min(cm), max(cm), sum(cm)

        lm = 253
        light_max = 80
        def to_255(light_val):
            return round(
                    lm + (light_max - lm) * (light_val - m) / (max_val - m)
                )  # light_val=m -> lm; light_val=max_val->light_max
        def to_color(light_val):
            return tuple([to_255(light_val)] * 3)

        for r in range(len(dataframe.index.tolist())):
            for c in range(len(dataframe.columns.tolist())):
                count = cmatrix[r][c]
                col = to_color(count)
                self.draw_rectangle(
                    [c - 0.5, c + 0.5],
                    [r - 0.5, r + 0.5],
                    marker="sd",
                    color=col,
                    fill=True,
                )

        y_labels = list(set(range(len(dataframe.columns))))
        x_labels = list(set(range(len(dataframe.columns))))

        self.set_yreverse(True)
        self.set_xticks(x_labels, xlabels)
        self.set_yticks(y_labels, ylabels)
        self.set_ticks_color(color)
        self.set_ticks_style(style)
        self.set_axes_color("default")
        self.set_canvas_color("default")
        self.set_title("Heatmap")
        print(dataframe)

    def draw_image(self, path, marker=None, style=None, fast=False, grayscale=False):
        from PIL import Image

        path = ut.correct_path(path)
        if not ut.is_file(path):
            return
        image = Image.open(path)
        self._draw_image(
            image, marker=marker, style=style, grayscale=grayscale, fast=fast
        )

    ##############################################
    #######    Plotting Tools Utilities    #######
    ##############################################

    def check_orientation(self, orientation=None, default_index=0):
        default = self.default.orientation
        default_first_letter = [el[0] for el in default]
        orientation = (
            default[default_first_letter.index(orientation)]
            if orientation in default_first_letter
            else orientation
        )
        orientation = (
            default[default_index] if orientation not in default else orientation
        )
        return orientation

    def check_alignment(self, alignment=None):
        default = self.default.alignment[0:-1]
        default_first_letter = [el[0] for el in default]
        alignment = (
            default[default_first_letter.index(alignment)]
            if alignment in default_first_letter
            else alignment
        )
        alignment = default[1] if alignment not in default else alignment
        return alignment

    def _draw_image(self, image, marker=None, style=None, fast=False, grayscale=False):
        from PIL import ImageOps

        image = ImageOps.grayscale(image) if grayscale else image
        image = image.convert("RGB")
        size = ut.update_size(image.size, self.size)
        image = image.resize(size, resample=True)
        matrix = ut.image_to_matrix(image)
        self.set_xfrequency(0)
        self.set_yfrequency(0)
        self.draw_matrix(matrix, marker=marker, style=style, fast=fast)
        self.set_xlabel()
        self.set_ylabel()



#!/usr/bin/env python3

"""
Specific test to identify and highlight gaps in pie chart rendering
"""

import sys

sys.path.insert(0, "../src")

try:
    import plotext_plus as plt

    print("✓ plotext_plus imported successfully")

    # Get full terminal dimensions for maximum resolution testing
    terminal_width, terminal_height = plt.terminal_size()
    print(f"📏 Using full terminal dimensions: {terminal_width}x{terminal_height}")

    # Test pie chart with simple, equal segments to make gaps more visible
    print("\nTesting pie chart gap detection...")
    labels = ["A", "B", "C", "D"]
    values = [25, 25, 25, 25]  # Equal segments make gaps more obvious
    colors = ["red", "blue", "green", "orange"]

    plt.clear_figure()
    plt.plotsize(terminal_width, terminal_height - 5)  # Use full terminal size

    # Create a pie chart that should have no gaps
    plt.pie(labels, values, colors=colors, title="Gap Detection Test - Equal Segments")
    plt.show()

    print("\n" + "=" * 60)
    print("GAP ANALYSIS:")
    print("=" * 60)
    print("Look at the pie chart above and check for:")
    print("1. Black spaces between colored segments")
    print("2. Incomplete filling within segments")
    print("3. Jagged or uneven edges")
    print("4. Missing blocks in what should be solid areas")
    print()
    print("If you see any black gaps between the red, blue, green,")
    print("and orange segments, the algorithm needs improvement.")
    print("=" * 60)

    # Test with unequal segments
    print("\nTesting with unequal segments...")
    labels2 = ["Large", "Medium", "Small", "Tiny"]
    values2 = [50, 30, 15, 5]
    colors2 = ["red", "blue", "green", "orange"]

    plt.clear_figure()
    plt.plotsize(terminal_width, terminal_height - 5)  # Use full terminal size again
    plt.pie(
        labels2, values2, colors=colors2, title="Gap Detection Test - Unequal Segments"
    )
    plt.show()

    print("\n" + "=" * 60)
    print("UNEQUAL SEGMENTS ANALYSIS:")
    print("=" * 60)
    print("Check especially:")
    print("1. Small segments (like 'Tiny' at 5%) for complete filling")
    print("2. Boundaries between different-sized segments")
    print("3. Whether thin segments have gaps or missing blocks")
    print("=" * 60)

except ImportError as e:
    print(f"❌ Import error: {e}")
except Exception as e:
    print(f"❌ Error during testing: {e}")
    import traceback

    traceback.print_exc()



#!/usr/bin/env python3

"""
Comprehensive test suite for the new Plotext API
Tests both the new modern interface and backward compatibility
"""

import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

import plotext_plus as plt


def test_backward_compatibility():
    """Test that all existing plotext functionality still works"""
    plt.log_info("Testing backward compatibility...")

    # Test original API - should work exactly as before
    x = list(range(10))
    y = [i**2 for i in x]

    plt.clear_figure()
    plt.scatter(x, y, color="red")
    plt.title("Backward Compatibility Test")
    plt.xlabel("X Values")
    plt.ylabel("Y Values")
    plt.show()

    plt.log_success("Backward compatibility test passed!")
    print("\n" + "=" * 60 + "\n")


def test_new_oo_api():
    """Test the new object-oriented Chart API with method chaining"""
    plt.log_info("Testing new object-oriented API...")

    # Test method chaining
    x = list(range(10))
    y = [i**2 for i in x]

    chart = (
        plt.Chart(use_banners=True, banner_title="OO API Demo")
        .scatter(x, y, color="blue", label="Quadratic")
        .title("Object-Oriented API Test")
        .xlabel("X Values")
        .ylabel("Y² Values")
        .size(width=80, height=20)
        .show()
    )

    plt.log_success("Object-oriented API test passed!")
    print("\n" + "=" * 60 + "\n")


def test_quick_functions():
    """Test the quick convenience functions"""
    plt.log_info("Testing quick convenience functions...")

    # Enable banners for all quick functions
    plt.enable_banners(True, "Quick Functions Demo")

    # Test quick_scatter
    x = list(range(5))
    y = [i * 2 for i in x]
    plt.quick_scatter(
        x,
        y,
        title="Quick Scatter Plot",
        xlabel="Time",
        ylabel="Value",
        use_banners=True,
        banner_title="Quick Scatter Demo",
    )

    # Test quick_line
    x2 = list(range(8))
    y2 = [i**0.5 for i in x2]
    plt.quick_line(
        x2,
        y2,
        title="Quick Line Plot",
        xlabel="Input",
        ylabel="Square Root",
        use_banners=True,
        banner_title="Quick Line Demo",
    )

    # Test quick_bar
    labels = ["A", "B", "C", "D"]
    values = [10, 25, 15, 30]
    plt.quick_bar(
        labels,
        values,
        title="Quick Bar Chart",
        use_banners=True,
        banner_title="Quick Bar Demo",
    )

    plt.log_success("Quick functions test passed!")
    print("\n" + "=" * 60 + "\n")


def test_multiple_series():
    """Test plotting multiple data series on the same chart"""
    plt.log_info("Testing multiple data series...")

    x = list(range(10))
    y1 = [i**2 for i in x]
    y2 = [i * 3 for i in x]
    y3 = [20 - i for i in x]

    chart = (
        plt.Chart(use_banners=True, banner_title="Multi-Series Chart")
        .scatter(x, y1, color="red", label="Quadratic")
        .line(x, y2, color="blue", label="Linear")
        .scatter(x, y3, color="green", label="Inverse")
        .title("Multiple Data Series")
        .xlabel("X Values")
        .ylabel("Y Values")
        .show()
    )

    plt.log_success("Multiple series test passed!")
    print("\n" + "=" * 60 + "\n")


def test_chuk_term_integration():
    """Test chuk-term output integration"""
    plt.log_info("Testing chuk-term output integration...")

    # Test all output types
    plt.log_info("This is an informational message")
    plt.log_success("This indicates successful completion")
    plt.log_warning("This is a warning message")
    plt.log_error("This indicates an error condition")

    plt.log_success("Chuk-term integration test passed!")
    print("\n" + "=" * 60 + "\n")


def test_banner_modes():
    """Test different banner modes and configurations"""
    plt.log_info("Testing banner mode configurations...")

    # Test with custom banner titles
    x = list(range(5))
    y = [i**3 for i in x]

    chart1 = (
        plt.Chart(use_banners=True, banner_title="🚀 Cubic Growth")
        .scatter(x, y, color="magenta")
        .title("Banner Mode: Custom Title")
        .show()
    )

    # Test banner title change
    chart2 = plt.Chart(use_banners=True, banner_title="📈 Dynamic Analysis")
    chart2.scatter(x, y, color="cyan")
    chart2.title("Banner Mode: Dynamic Title")
    chart2.show()

    plt.log_success("Banner modes test passed!")
    print("\n" + "=" * 60 + "\n")


def run_all_tests():
    """Run all test suites"""
    plt.log_info("🧪 Starting comprehensive Plotext API tests...\n")

    try:
        test_backward_compatibility()
        test_new_oo_api()
        test_quick_functions()
        test_multiple_series()
        test_chuk_term_integration()
        test_banner_modes()

        plt.log_success("🎉 All tests passed successfully!")
        plt.log_info("✨ New API is ready for use!")

    except Exception as e:
        plt.log_error(f"❌ Test failed: {str(e)}")
        raise


if __name__ == "__main__":
    run_all_tests()



#!/usr/bin/env python3
"""
Advanced Plotext features demonstration.
Shows banner customization, themes, and complex visualizations.
"""

import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

import math
import random

import plotext_plus as plt


def custom_banner_example():
    """Demonstrate custom banner titles and styling"""
    plt.log_info("🎨 Demonstrating custom banners...")

    x = list(range(25))
    y = [10 + 5 * math.sin(i / 3) + random.randint(-2, 2) for i in x]

    # Custom themed banner
    (
        plt.Chart(use_banners=True, banner_title="🌊 Signal Processing")
        .line(x, y, color="cyan", label="Noisy Signal")
        .title("Real-time Data Analysis")
        .xlabel("Time Samples")
        .ylabel("Amplitude")
        .show()
    )


def scientific_visualization():
    """Scientific data visualization example"""
    plt.log_info("🔬 Creating scientific visualization...")

    # Generate experimental data
    temperature = list(range(20, 101, 5))
    pressure = [t * 0.8 + random.randint(-5, 5) for t in temperature]
    theoretical = [t * 0.75 for t in temperature]

    chart = (
        plt.Chart(use_banners=True, banner_title="🧪 Laboratory Results")
        .scatter(temperature, pressure, color="red", label="Experimental")
        .line(temperature, theoretical, color="blue", label="Theoretical")
        .title("Temperature vs Pressure Analysis")
        .xlabel("Temperature (°C)")
        .ylabel("Pressure (kPa)")
        .show()
    )


def business_dashboard():
    """Business dashboard example"""
    plt.log_info("📊 Creating business dashboard...")

    # Revenue data
    months = ["Jan", "Feb", "Mar", "Apr", "May", "Jun"]
    revenue = [120, 135, 158, 142, 167, 181]
    target = [150] * len(months)

    chart = (
        plt.Chart(use_banners=True, banner_title="💼 Revenue Dashboard")
        .bar(months, revenue, color="green")
        .line(list(range(len(months))), target, color="red", label="Target")
        .title("Monthly Revenue Performance")
        .show()
    )


def mathematical_functions():
    """Mathematical function plotting"""
    plt.log_info("📐 Plotting mathematical functions...")

    x = [i / 10 for i in range(-50, 51)]
    sin_values = [math.sin(val) for val in x]
    cos_values = [math.cos(val) for val in x]

    chart = (
        plt.Chart(use_banners=True, banner_title="📈 Mathematical Functions")
        .line(x, sin_values, color="blue", label="sin(x)")
        .line(x, cos_values, color="red", label="cos(x)")
        .title("Trigonometric Functions")
        .xlabel("x")
        .ylabel("f(x)")
        .show()
    )


def performance_comparison():
    """Performance comparison chart"""
    plt.log_info("⚡ Creating performance comparison...")

    algorithms = ["Algorithm A", "Algorithm B", "Algorithm C", "Algorithm D"]
    execution_times = [2.3, 1.8, 3.1, 1.2]
    memory_usage = [45, 32, 67, 28]

    # Time comparison
    plt.quick_bar(
        algorithms,
        execution_times,
        title="Execution Time Comparison",
        use_banners=True,
        banner_title="⏱️ Performance Metrics",
    )

    print("\n" + "-" * 40 + "\n")

    # Memory comparison
    plt.quick_bar(
        algorithms,
        memory_usage,
        title="Memory Usage Comparison",
        use_banners=True,
        banner_title="💾 Memory Analysis",
    )


def output_features_demo():
    """Demonstrate chuk-term output features"""
    plt.log_info("🎯 Demonstrating output features...")

    # Various output levels
    plt.log_info("📋 Processing data...")
    plt.log_success("✅ Data loaded successfully")
    plt.log_warning("⚠️ Using default parameters")
    plt.log_error("❌ Minor calculation error detected")

    # Simple visualization with context
    x = list(range(8))
    y = [i**1.5 for i in x]

    plt.log_info("📊 Generating final visualization...")

    chart = (
        plt.Chart(use_banners=True, banner_title="🎯 Final Results")
        .scatter(x, y, color="purple", label="Growth Curve")
        .title("Project Completion Analysis")
        .xlabel("Weeks")
        .ylabel("Progress")
        .show()
    )

    plt.log_success("🎉 Analysis complete!")


def main():
    """Run all advanced examples"""
    plt.log_success("🌟 Starting Advanced Plotext Features Demo")

    custom_banner_example()
    print("\n" + "=" * 60 + "\n")

    scientific_visualization()
    print("\n" + "=" * 60 + "\n")

    business_dashboard()
    print("\n" + "=" * 60 + "\n")

    mathematical_functions()
    print("\n" + "=" * 60 + "\n")

    performance_comparison()
    print("\n" + "=" * 60 + "\n")

    output_features_demo()

    plt.log_success("✨ All advanced examples completed!")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
import os
import sys

import plotext_plus as plt
from plotext_plus._utility import all_markers
from plotext_plus._utility import colors

try:
    import shtab
except ImportError:
    from . import _shtab as shtab


# For Possible Colors and Markers Completion
def dict_to_complete(d=None):
    if d is None:
        d = {}
    return {
        "zsh": "(("
        + " ".join(str(x[0]) + "\\:" + x[1] for x in d.items())
        + "))"
    }


def list_to_complete(data):
    return {"zsh": "((" + " ".join([el + "\\:" for el in data]) + "))"}


def build_parser():

    examples = """Access each function documentation for further guidance, eg: plotext scatter -h"""

    parser = argparse.ArgumentParser(
        prog="plotext",
        description="plotting directly on terminal",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    shtab.add_argument_to(parser, ["-s", "--print-completion"])

    parser.set_defaults(type="scatter")

    path_parser = argparse.ArgumentParser(add_help=False)

    path_parser.add_argument(
        "-p",
        "--path",
        action="store",
        # dest   = 'path',
        type=str,
        metavar="FILE",
        help="file path of the data table; if not used it will read from stdin. Use 'test' to automatically download, in your user folder, some test data/image/gif or video, depending on the function used; the file will be removed after the plot",
    ).complete = shtab.FILE
    path_parser.add_argument(
        "-r",
        "--first-row",
        action="store",
        type=int,
        default=0,
        metavar="FIRST-ROW",
        help="The first line to consider in the data file (counting from 0).",
    )

    common_parser = argparse.ArgumentParser(add_help=False)

    common_parser.add_argument(
        "-clt",
        "--clear_terminal",
        nargs=1,
        type=str,
        default=["False"],
        choices=["True", "False"],
        metavar="BOOL",
        help="it clears the terminal before plotting, if True (as by default)",
    )

    common_parser.add_argument(
        "-s",
        "--sleep",
        nargs=1,
        type=float,
        default=[0],
        help="it adds a sleeping time after plotting, to reduce flickering when multiple plots are required; 0 by default",
    )

    data_parser = argparse.ArgumentParser(add_help=False)

    data_parser.add_argument(
        "-xcol",
        "--xcolumn",
        nargs="+",  # 1 or more
        type=str,
        default=["none"],
        help="the column number (starting from 1), in the data table, that will be used as x data; by default 'none'",
    )

    data_parser.add_argument(
        "-ycols",
        "--ycolumns",
        nargs="+",  # 1 or more
        type=str,
        default=["all"],
        help="the column numbers (starting from 1), in the data table, that will be used as y data; by default 'all'",
    )

    data_parser.add_argument(
        "-d",
        "--delimiter",
        type=str,
        default=[" "],
        nargs=1,
        help="character to be used to separate columns in the data table (eg: ;); by default the white space ' '",
    )

    data_parser.add_argument(
        "-l",
        "--lines",
        nargs="+",  # 1 or more
        type=int,
        default=[1000],
        help="number of lines, from data table, to be plotted at each iteration; 1000 by default; data shorter then this will be plotted in a single iteration",
    )

    options_parser = argparse.ArgumentParser(add_help=False)

    options_parser.add_argument(
        "-m",
        "--marker",
        type=str,
        default=["hd"],
        nargs=1,
        choices=all_markers.keys(),
        metavar="marker",
        help="character or marker code identifying the data points in the plot, eg: x, braille, sd, heart, fhd; by default hd",
    ).complete = dict_to_complete(all_markers)

    options_parser.add_argument(
        "-c",
        "--color",
        type=str,
        default=[None],
        nargs=1,
        choices=colors,
        metavar="COLOR",
        help="color of the data points in the plot, between: "
        + ",".join(colors)
        + "; add + at the end of the color (except black and white) for clearer colors, eg: red+",
    ).complete = list_to_complete(colors)

    options_parser.add_argument(
        "-t", "--title", nargs=1, type=str, default=[None], help="the plot title"
    )

    options_parser.add_argument(
        "-xl",
        "--xlabel",
        nargs=1,
        type=str,
        default=[None],
        help="the label of the x axis",
    )

    options_parser.add_argument(
        "-yl",
        "--ylabel",
        nargs=1,
        type=str,
        default=[None],
        help="the label of the y axis",
    )

    options_parser.add_argument(
        "-g",
        "--grid",
        nargs=1,
        type=str,
        default=["False"],
        choices=["True", "False"],
        metavar="BOOL",
        help="it add grid lines if True, or removed them if False (as by default)",
    )

    barhist_parser = argparse.ArgumentParser(add_help=False)

    barhist_parser.add_argument(
        "-o",
        "--orientation",
        nargs=1,
        type=str,
        default=["v"],
        choices=["v", "h"],
        metavar="ORIENTATION",
        help="bar orientation, v for vertical (as by default), h for horizontal",
    )

    barhist_parser.add_argument(
        "-f",
        "--fill",
        nargs=1,
        type=str,
        default=["True"],
        choices=["True", "False"],
        metavar="BOOL",
        help="it fills the bars with colored markers if True (as by default), otherwise only the bars skeleton is shown",
    )

    subparser = parser.add_subparsers(dest="type", help="the user defined plot type")

    subparser.add_parser(
        "scatter",
        description="plots a series of data points",
        parents=[path_parser, data_parser, common_parser, options_parser],
        help="plots a series of data points",
        epilog="eg: plotext scatter --path test --xcolumn 1 --ycolumns 2 --lines 5000 --title 'Scatter Plot Test' --marker braille",
    )

    subparser.add_parser(
        "plot",
        parents=[path_parser, data_parser, common_parser, options_parser],
        description="plots lines between consecutive data points",
        help="plots lines between consecutive data points",
        epilog="eg: plotext plot --path test --xcolumn 1 --ycolumns 2 --sleep 0.1 --lines 2500 --clear_terminal True --color magenta+ --title 'Plot Test'",
    )

    subparser.add_parser(
        "plotter",
        parents=[path_parser, data_parser, common_parser, options_parser],
        description="plots a series of data points and the lines between consecutive ones",
        help="scatter + plot",
        epilog="eg:plotext plotter --path test --xcolumn 1 --ycolumns 2 --sleep 0.1 --lines 120 --clear_terminal True --marker hd --title 'Plotter Test'",
    )

    bar = subparser.add_parser(
        "bar",
        parents=[
            path_parser,
            data_parser,
            common_parser,
            options_parser,
            barhist_parser,
        ],
        description="builds a bar plot",
        help="bar plot",
        epilog="eg: plotext bar --path test --xcolumn 1 --title 'Bar Plot Test' --xlabel Animals --ylabel Count",
    )

    bar.add_argument(
        "-w",
        "--width",
        nargs=1,
        type=float,
        default=[None],
        help="bars width as a float between 0 and 1",
    )

    hist = subparser.add_parser(
        "hist",
        parents=[
            path_parser,
            data_parser,
            common_parser,
            options_parser,
            barhist_parser,
        ],
        description="builds a histogram plot",
        help="histogram plot",
        epilog="eg: plotext hist --path test --xcolumn 1 --ycolumns 2 --lines 5000 --title 'Histogram Test'",
    )

    hist.add_argument(
        "-b",
        "--bins",
        nargs=1,
        type=int,
        default=[10],
        help="histogram bins (10 by default)",
    )

    subparser.add_parser(
        "image",
        parents=[path_parser, common_parser],
        description="plots an image from path",
        help="plots an image from file path",
        epilog="eg: plotext image --path test",
    )

    subparser.add_parser(
        "gif",
        parents=[path_parser, common_parser],
        description="plays a gif image from path",
        help="plays a gif image from path",
        epilog="eg: plotext gif --path test",
    )

    video = subparser.add_parser(
        "video",
        parents=[path_parser, common_parser],
        description="plays a video from path",
        help="plays a video from path",
        epilog="eg: plotext video --path test --from_youtube True",
    )

    video.add_argument(
        "-fy",
        "--from_youtube",
        nargs=1,
        type=str,
        default=["False"],
        choices=["True", "False"],
        metavar="BOOL",
        help="set it to True to render the colors correctly for videos downloaded from youtube; default is False",
    )

    youtube = subparser.add_parser(
        "youtube",
        description="plays a youtube video from url",
        help="plays a youtube video from url",
        epilog="eg: plotext youtube --url test",
    )

    youtube.add_argument(
        "-u",
        "--url",
        action="store",
        dest="url",
        nargs=1,
        type=str,
        metavar="URL",
        help="the url of a youtube video; use 'test' for a test video",
    )

    return parser


def main(argv=None):
    parser = build_parser()
    args = parser.parse_args(argv)

    type = args.type
    first_row = 0
    if type != "youtube":
        # Handle case when path is not provided (no subcommand)
        if not hasattr(args, "path"):
            print(
                "Error: No subcommand provided. Use --help to see available commands."
            )
            print(
                "Available commands: scatter, plot, plotter, bar, hist, image, gif, video, youtube"
            )
            print("Example: plotext_plus scatter --path test")
            return

        path = args.path
        first_row = args.first_row
        clt = args.clear_terminal[-1] == "True"
        sleep = args.sleep[0]

    def get_x_y(data):
        data_length = len(data)
        xcol = args.xcolumn[0]
        ycols = args.ycolumns
        xcol = (
            "none"
            if xcol == "none"
            else int(xcol) if int(xcol) - 1 in range(data_length) else "none"
        )
        all_ycols = [el for el in range(data_length) if el != xcol]
        ycols = (
            all_ycols
            if ycols == ["all"]
            else [int(el) for el in ycols if int(el) - 1 in range(data_length)]
        )
        x = list(range(1, len(data[0]) + 1)) if xcol == "none" else data[xcol - 1]
        y_values = [data[i - 1] for i in ycols]
        return x, y_values

    def plot(x, y_values):
        for y in y_values:
            (
                plt.plot(x, y, marker=marker, color=color)
                if type in ["plot", "plotter"]
                else None
            )
            (
                plt.scatter(
                    x, y, marker=marker, color=None if type == "plotter" else color
                )
                if type in ["scatter", "plotter"]
                else None
            )
            (
                plt.bar(
                    x,
                    y,
                    marker=marker,
                    color=color,
                    width=width,
                    orientation=orientation,
                    fill=fill,
                )
                if type == "bar"
                else None
            )
            (
                plt.hist(
                    y,
                    marker=marker,
                    color=color,
                    orientation=orientation,
                    fill=fill,
                    bins=bins,
                )
                if type == "hist"
                else None
            )
        plt.clt() if clt else None
        plt.show()
        plt.cld()
        plt.sleep(sleep)

    data_plot = ["scatter", "plot", "plotter", "bar", "hist"]
    test_path = "test"

    if type in data_plot:
        lines = args.lines[0]
        delimiter = args.delimiter[0]
        title = args.title[0]
        xlabel = args.xlabel[0]
        ylabel = args.ylabel[0]
        grid = args.grid[-1] == "True"

        orientation = args.orientation[0] if type in ["bar", "hist"] else "v"
        fill = args.fill[0] == "True" if type in ["bar", "hist"] else False
        width = args.width[0] if type == "bar" else 0
        bins = args.bins[0] if type in "hist" else None

        marker = args.marker[0]
        color = args.color[0]

        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.grid(grid)

        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            if type != "bar":
                plt.download(plt.test_data_url, test_path, log=True)
            else:
                plt.download(plt.test_bar_data_url, test_path, log=True)
            path = test_path

        if path is None:

            def plot_text(text):
                data = plt._utility.read_lines(text, delimiter=delimiter)
                data = plt.transpose(data)
                x, y_values = get_x_y(data)
                plot(x, y_values)

            for _ in range(first_row):
                sys.stdin.readline()
            text = []
            for line in iter(sys.stdin.readline, ""):
                text.append(line)
                if len(text) == lines:
                    plot_text(text)
                    text = []
            if (
                len(text) > 0
            ):  # this is when there is some residual data to be plotted, not lines long
                plot_text(text)
                text = []

        else:
            data = plt.read_data(path, delimiter=delimiter, first_row=first_row)
            data = plt.transpose(data)
            x, y_values = get_x_y(data)
            chunks = len(x) // lines + (1 if len(x) % lines else 0)
            for c in range(chunks):
                xc = x[c * lines : (c + 1) * lines]
                y_chunk = [y[c * lines : (c + 1) * lines] for y in y_values]
                plot(xc, y_chunk)

    elif type == "image":
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_image_url, test_path, log=True)
            path = test_path
        plt.image_plot(path, fast=True)
        plt.clt() if clt else None
        plt.show()

    elif type == "gif":
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_gif_url, test_path, log=True)
            path = test_path
        plt.play_gif(path)

    elif type == "video":
        if path == test_path:
            plt.plotsize(None, plt.th() - 3)
            plt.download(plt.test_video_url, test_path, log=True)
            path = test_path
        from_youtube = args.from_youtube[-1] == "True"
        plt.play_video(path, from_youtube)

    elif type == "youtube":
        url = args.url[-1]
        url = plt.test_youtube_url if url == test_path else url
        plt.play_youtube(url)

    if os.path.isfile(test_path):
        plt.delete_file(test_path, True)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

"""
Plotext Plus MCP Server CLI
==========================

Command-line interface for starting the Plotext Plus MCP server.
"""

import argparse
import sys


def main():
    """Main entry point for the MCP server CLI."""
    parser = argparse.ArgumentParser(
        description="Plotext Plus MCP Server - Terminal plotting for AI clients"
    )
    parser.add_argument(
        "--version", action="version", version="plotext_plus MCP server"
    )
    parser.add_argument(
        "--info", action="store_true", help="Show MCP server information"
    )
    parser.add_argument(
        "--stdio",
        action="store_true",
        help="Use STDIO transport mode instead of HTTP (requires dev version of chuk-mcp-server)",
    )

    args = parser.parse_args()

    if args.info:
        print("Plotext Plus MCP Server")
        print("======================")
        print("Terminal plotting library with MCP (Model Context Protocol) support")
        print("Exposes plotting, charting, theming, and utility functions as MCP tools")
        print("")
        print("Available tools:")
        print("- scatter_plot: Create scatter plots")
        print("- line_plot: Create line plots")
        print("- bar_chart: Create bar charts")
        print("- matrix_plot: Create matrix/heatmap plots")
        print("- quick_scatter/line/bar: Quick chart creation")
        print("- theme management tools")
        print("- utility functions")
        print("")
        print("Available prompts:")
        print("- basic_scatter: Simple scatter plot example")
        print("- basic_bar_chart: Bar chart example")
        print("- multi_step_workflow: Complex multi-step analysis")
        print("- theme_exploration: Theme comparison examples")
        print("- regional_sales_analysis: Interactive data analysis")
        print("- performance_testing: Large dataset testing")
        print("- error_handling_test: Edge case testing")
        print("- complete_workflow: End-to-end visualization workflow")
        print("- And 7 more prompts covering all use cases")
        print("")
        print("Available resources:")
        print("- config://plotext: Server configuration and capabilities")
        print("")
        print("To start the server, run: plotext-mcp")
        print("For STDIO transport mode, run: plotext-mcp --stdio")
        return

    try:
        from .mcp_server import start_server

        start_server(stdio_mode=args.stdio)
    except ImportError as e:
        if "chuk-mcp-server" in str(e):
            print("ERROR: MCP functionality requires chuk-mcp-server")
            print("Install it with: uv add --optional mcp plotext_plus")
            print("Or: pip install 'plotext_plus[mcp]'")
            sys.exit(1)
        else:
            raise


if __name__ == "__main__":
    main()



# This file contains some plotext functions which are only available to the top main level and not to sub figures (which are written in _figure.py and _monitor.py). These are functions which requires some coding and would be too long to be added directly in _core.py

import datetime as dt
from math import ceil
from math import sqrt
from time import sleep
from time import time

import plotext_plus._utility as ut
from plotext_plus._figure import _FigureClass
from plotext_plus._utility import hd_symbols
from plotext_plus._utility import marker_codes
from plotext_plus._utility import themes as _themes

figure = _FigureClass()  # the main figure at top level

##############################################
#######     Simple  Bar Functions    ########
##############################################


def simple_bar(*args, width=None, marker=None, color=None, title=None):
    x, y = ut.set_data(*args)
    marker = ut.correct_marker(marker)

    color_ok = ut.is_color(color) or (isinstance(color, list) and len(color) == len(x))
    color = [color] if color_ok else None

    simple_stacked_bar(x, [y], width=width, marker=marker, colors=color, title=title)


def simple_stacked_bar(
    *args, width=None, marker=None, colors=None, title=None, labels=None
):
    x, y, y_scaled, width = ut.bar_data(*args, width=width)
    marker = ut.correct_marker(marker)

    bars = len(y_scaled)
    stacked_bars = len(y_scaled[0])

    colors_ok1 = (
        isinstance(colors, list)
        and isinstance(colors[0], list)
        and ut.matrix_size(colors) == [bars, stacked_bars]
    )
    colors_ok2 = isinstance(colors, list) and len(colors) == stacked_bars
    colors = (
        ut.transpose(colors)
        if colors_ok1
        else (
            [colors] * bars if colors_ok2 else [ut.color_sequence[:stacked_bars]] * bars
        )
    )

    title = ut.get_title(title, width)
    bars = [ut.single_bar(x[i], y_scaled[i], y[i], marker, colors[i]) for i in range(bars)]
    labels = ut.get_simple_labels(marker, labels, colors[0], width)
    figure.monitor.matrix.canvas = title + "\n".join(bars) + labels
    figure.monitor.fast_plot = True


def simple_multiple_bar(
    *args, width=None, marker=None, colors=None, title=None, labels=None
):
    x, y, y_scaled, width = ut.bar_data(*args, width=width, mode="multiple")
    bars = len(y_scaled)
    multiple_bars = len(y_scaled[0])
    lx = len(x[0])
    marker = ut.correct_marker(marker)

    colors_ok = isinstance(colors, list) and len(colors) == multiple_bars
    colors = colors if colors_ok else ut.color_sequence[:multiple_bars]

    out = ut.get_title(title, width)
    for i in range(bars):
        xn = [
            x[i] if j == (multiple_bars - 1) // 2 else ut.space * lx
            for j in range(multiple_bars)
        ]
        new = [
            ut.single_bar(xn[j], [y_scaled[i][j]], y[j][i], marker, [colors[j]])
            for j in range(multiple_bars)
        ]
        out += "\n".join(new)
        out += "\n\n" if i != bars - 1 else ""
    labels = ut.get_simple_labels(marker, labels, colors, width)
    figure.monitor.matrix.canvas = out + labels
    figure.monitor.fast_plot = True


##############################################
#############     Play GIF    ################
##############################################


def play_gif(path):
    from PIL import Image
    from PIL import ImageSequence

    path = ut.correct_path(path)
    if not ut.is_file(path):
        return
    im = Image.open(path)
    for image in ImageSequence.Iterator(im):
        load_time = time()
        figure.clt()
        image = image.convert("RGB")
        figure.monitor._draw_image(image, fast=True)
        figure.show()
        load_time = time() - load_time
        frame_time = image.info["duration"] / 10**3
        if load_time < frame_time:
            sleep(frame_time - load_time)


##############################################
##########     Video Functions    ############
##############################################


def play_video(path, from_youtube=False):
    path = ut.correct_path(path)
    if not ut.is_file(path):
        return
    _play_video(path, from_youtube)


def play_youtube(url):
    import pafy

    video = pafy.new(url)
    best = video.getbest()
    _play_video(best.url, from_youtube=True)


def get_youtube(url, path, log):
    import pafy

    video = pafy.new(url)
    best = video.getbest(preftype="mp4")
    path = "youtube-video.mp4" if path is None else path
    path = ut.correct_path(path)
    best.download(filepath=path, quiet=not log)
    print(ut.format_strings("YouTube video downloaded as", path)) if log else None


def _play_video(path, from_youtube=False):
    import cv2
    from ffpyplayer.player import MediaPlayer
    from PIL import Image

    cap = cv2.VideoCapture(path)
    player = MediaPlayer(path)  # , paused = True, loglevel = 'quiet');
    fr = 0
    while fr == 0:
        fr = cap.get(cv2.CAP_PROP_FPS)
    frame_time = 1 / fr
    # to_list = lambda frame: [[tuple(int(el) for el in tup) for tup in row] for row in frame]
    def pt(time):
        return f"{round(10 ** 3 * time, 1):05.1f}  "
    real_time = video_time = 0
    while True:
        load_time = time()
        check_video, frame = cap.read()
        audio, check_audio = player.get_frame(show=False)
        load_time = time() - load_time
        if not check_video:
            break
        if load_time >= frame_time:
            continue
        real_time += load_time
        video_time += frame_time
        show_time = 0
        if video_time >= real_time:
            show_time = time()
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB) if from_youtube else frame
            # frame = to_list(frame)
            image = Image.fromarray(frame)

            # Synchronized frame rendering - ensure each step completes before next
            figure.clt()
            # Small delay to ensure clear operation completes
            sleep(0.001)  # 1ms delay
            figure.monitor._draw_image(image, fast=True)
            # Small delay to ensure image drawing completes before display
            sleep(0.001)  # 1ms delay
            figure.show()
            # Brief delay after display to prevent frame overlap
            sleep(0.002)  # 2ms delay for terminal rendering

            show_time = time() - show_time
        sleep_time = 0
        if real_time < video_time:
            sleep_time = time()
            sleep(video_time - real_time)
            sleep_time = time() - sleep_time
        load_time + show_time + sleep_time
        real_time += show_time + sleep_time
        # print('load: ' + pt(load_time), 'show: ' + pt(show_time), 'sleep: ' + pt(sleep_time), 'total: ' + pt(total_time), 'frame: ' + pt(frame_time), 'real: ' + pt(real_time), 'video: ' + pt(video_time), 'r/v:', round(real_time / video_time, 3)) if shown else None
    player.close_player()
    cap.release()
    cv2.destroyAllWindows()
    figure.clf()


##############################################
############     Utilities     ###############
##############################################

test_data_url = (
    "https://raw.githubusercontent.com/piccolomo/plotext/master/data/data.txt"
)
test_bar_data_url = (
    "https://raw.githubusercontent.com/piccolomo/plotext/master/data/bar_data.txt"
)
test_image_url = (
    "https://raw.githubusercontent.com/piccolomo/plotext/master/data/cat.jpg"
)
test_gif_url = (
    "https://raw.githubusercontent.com/piccolomo/plotext/master/data/homer.gif"
)
test_video_url = (
    "https://raw.githubusercontent.com/piccolomo/plotext/master/data/moonwalk.mp4"
)
test_youtube_url = "https://www.youtube.com/watch?v=ZNAvVVc4b3E&t=75s"

##############################################
#########     Matplotlib Backend    ##########
##############################################


def from_matplotlib(fig, marker=None):
    fig.canvas.draw()
    slots = (rows, cols) = fig.axes[0].get_subplotspec().get_gridspec().get_geometry()
    figure.clf()  # clt()
    figure.subplots(*slots)
    def round10(data):
        return [round(el, 10) for el in data]
    def to_rgb(rgb_norm):
        return tuple([round(255 * el) for el in rgb_norm[:3]])
    figure.axes_color(to_rgb(fig.patch.get_facecolor()))
    for sub in fig.axes[:]:
        p = sub.get_subplotspec().get_geometry()[2]
        row = int((p - 0) / cols + 1)
        col = p + 1 - (row - 1) * cols
        monitor = figure.subplot(row, col)
        monitor.xlabel(sub.get_xlabel())
        monitor.ylabel(sub.get_ylabel())
        monitor.title(sub.get_title())
        monitor.xscale(sub.get_xscale())
        monitor.yscale(sub.get_yscale())
        monitor.xticks(round10(sub.get_xticks()))
        monitor.yticks(round10(sub.get_yticks()))
        monitor.canvas_color(to_rgb(sub.get_facecolor()))
        for point in sub.collections:
            label = point.get_label()
            label = label if label[0] != "_" else ""
            # point.set_offset_position('data')
            x, y = ut.transpose(point.get_offsets())
            color = [ut.to_rgb(point.to_rgba(el)) for el in point.get_facecolors()[0]]
            # can't find the right point colors
            monitor.scatter(x, y, label=label, marker=marker)
        for line in sub.get_lines():
            label = line.get_label()
            label = label if label[0] != "_" else ""
            x, y = line.get_data()
            monitor.plot(x, y, marker=marker, color=line.get_c(), label=label)
        for b in sub.patches:
            label = b.get_label()
            label = label if label[0] != "_" else ""
            color = b.get_facecolor()
            color = ut.to_rgb(color)
            box = b.get_bbox()
            x0, y0, x1, y1 = box.x0, box.y0, box.x1, box.y1
            x = [x0, x0, x1, x1, x0]
            y = [y0, y1, y1, y0, y0]
            fill = b.get_fill()
            fillx = fill if y0 == 0 else False
            filly = fill if x0 == 0 else False
            monitor.plot(
                x, y, fillx=fillx, filly=filly, marker=marker, color=color, label=label
            )
        monitor.xlim(*sub.get_xlim())
        monitor.ylim(*sub.get_ylim())


##############################################
#######     Presentation Functions    ########
##############################################


def markers():
    markers = list(hd_symbols.keys())[::-1] + list(marker_codes.keys())
    marker_count = len(markers)
    rows = int(sqrt(marker_count))
    cols = ceil(marker_count / rows)
    y = ut.sin(1)
    figure.clf()
    figure.theme("pro")
    figure.xfrequency(0)
    figure.yfrequency(0)
    figure.frame(1)
    figure.subplots(rows, cols)
    figure.frame(0)
    for row in range(1, rows + 1):
        for col in range(1, cols + 1):
            i = (row - 1) * cols + col - 1
            if i < marker_count:
                subplot = figure.subplot(row, col)
                figure.frame(1)
                default = " [default]" if markers[i] == "hd" else ""
                subplot.title(markers[i] + default)
                subplot.scatter(y, marker=markers[i])
                subplot.ticks_style("bold")
                # figure.ticks_color(figure._utility.title_color)
    figure.show()
    figure.clf()


def colors():
    print(ut.colorize("String Color Codes", style="bold"))
    bg = "default"
    c = ut.no_duplicates(
        [
            el.replace("+", "")
            for el in ut.colors
            if el not in ["default", "black", "white"]
        ]
    )
    cp = [ut.colorize(ut.pad_string(el + "+", 10), el + "+", background=bg) for el in c]
    c = [ut.colorize(ut.pad_string(el, 10), el, background=bg) for el in c]
    c = [" " + c[i] + cp[i] for i in range(len(c))]
    c = "\n".join(c)
    print(" " + ut.colorize(ut.pad_string("default", 20), background=bg))
    print(
        " "
        + ut.colorize(ut.pad_string("black", 10), "black", background="gray")
        + ut.colorize(ut.pad_string("white", 10), "white", background=bg)
    )
    print(c)
    print()
    # print(colorize("\n\nInteger Color Codes:", style = ''))
    c = ut.colorize("Integer Color Codes", style="bold", show=False) + "\n"
    for row in range(16):
        cr = " "
        for col in range(16):
            i = row * 16 + col
            cr += ut.colorize(ut.pad_string(i, 5), i)
        c += cr + "\n"
    print(c)
    c = "\n"
    rgb = (100, 200, 85)
    rgb_string = "(" + ", ".join([str(el) for el in rgb]) + ")"
    print(
        ut.colorize("RGB Tuples like:", style="bold"),
        ut.colorize(rgb_string, rgb, "bold"),
    )


def styles():
    from plotext_plus._utility import colorize
    from plotext_plus._utility import styles
    from plotext_plus._utility import title_color

    c = [colorize(el, style=el) for el in styles]
    c = "\n".join(c)
    print(c)
    mul = "bold italic dim"
    print(
        "\n"
        + colorize("multiple styles are accepted, ", title_color)
        + "eg: "
        + colorize(mul, style=mul)
    )


def themes():
    themes = list(_themes.keys())[::]
    theme_count = len(themes)
    rows = int(sqrt(theme_count))
    cols = ceil(theme_count / rows)
    y1 = ut.sin(periods=1)
    y2 = ut.sin(periods=1, phase=-1)
    figure.clf()
    figure.subplots(rows, cols)
    for row in range(1, rows + 1):
        for col in range(1, cols + 1):
            i = (row - 1) * cols + col - 1
            if i < theme_count:
                subplot = figure.subplot(row, col)
                subplot.theme(themes[i])
                subplot.title(themes[i])
                subplot.scatter(y1)
                subplot.plot(y2)
    figure.show()
    figure.clf()


##############################################
###########     Test Function    #############
##############################################


def test():
    import random

    figure.clf()
    figure.clt()
    figure.date_form("d/m/Y")
    figure.take_min()
    figure.plot_size(None, ut.terminal_height())
    # figure.plot_size(108, 70)

    figure.plotsize(ut.tw(), ut.th() - 5)
    figure.subplots(2, 2)

    subplot = figure.subplot(1, 1)
    subplot.title("Multiple Axes Plot")
    subplot.canvas_color(66)
    subplot.axes_color(4)
    subplot.ticks_color(216)
    subplot.ticks_style("bold italic")
    y = ut.sin(periods=1)
    data_length = len(y)
    subplot.scatter(y, label="lower left")
    x = [figure.today_datetime() + dt.timedelta(days=i) for i in range(data_length)]
    x = figure.datetimes_to_strings(x)
    subplot.plot(x, x, label="upper right - all dates", xside=2, yside=2)
    subplot.vline(data_length / 2, "red")
    subplot.hline(0, 200)
    subplot.text("origin", data_length // 2, 0, color="red", alignment="center")
    subplot.xlabel("x lower")
    subplot.xlabel("x upper", 2)
    subplot.ylabel("y left", "left")
    subplot.ylabel("y right", "right")
    subplot.xfrequency(8)
    subplot.xfrequency(5, 2)
    subplot.yfrequency(3)
    subplot.yfrequency(5, 2)
    subplot.grid(1, 1)

    subplot = figure.subplot(1, 2)
    subplot.theme("innocent")
    xb = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
    y1 = [36, 14, 11, 8, 7, 4]
    y2 = [20, 12, 35, 15, 4, 5]
    subplot.stacked_bar(xb, [y1, y2], labels=["men", "women"])

    subplot = figure.subplot(2, 1)
    subplot.theme("dreamland")
    ld = 7 * 10**4
    data = [random.gauss(0, 1) for el in range(10 * ld)]
    subplot.hist(data, bins=60, label="mean 0")
    subplot.frame(1)  # subplot.xaxes(1, 0); subplot.yaxes(1, 0)

    subplot = figure.subplot(2, 2)
    subplot.canvas_color("gray+")
    subplot.axes_color("gray+")
    ut.download(test_image_url, "cat.jpg")
    subplot.image_plot("cat.jpg", grayscale=False)
    ut.delete_file("cat.jpg")
    subplot.title("A very Cute Cat")
    subplot.frame(0)

    # figure.plotsize(0, 0)
    figure.show()
    figure._get_time()
    figure.save_fig("test.txt")
    figure.save_fig("test.html")
    # figure.clf()



#!/usr/bin/env python3

"""
Test to identify unfilled areas WITHIN pie chart segments
"""

import sys

sys.path.insert(0, "../src")

try:
    import plotext_plus as plt

    print("✓ plotext_plus imported successfully")

    # Test with a single large segment to make gaps obvious
    print("\nTesting single large segment for internal gaps...")
    labels = ["Single"]
    values = [100]
    colors = ["red"]

    plt.clear_figure()
    plt.pie(labels, values, colors=colors, title="Single Segment - Gap Detection")
    plt.show()

    print("\n" + "=" * 60)
    print("SINGLE SEGMENT ANALYSIS:")
    print("=" * 60)
    print("This should be a SOLID RED CIRCLE with NO black gaps inside.")
    print("Look carefully within the red area for:")
    print("1. Black spaces inside what should be solid red")
    print("2. Sparse/patchy areas with missing red blocks")
    print("3. Any non-red pixels within the circle boundary")
    print("=" * 60)

    # Test with two equal segments
    print("\nTesting two equal segments...")
    labels2 = ["Half A", "Half B"]
    values2 = [50, 50]
    colors2 = ["blue", "orange"]

    plt.clear_figure()
    plt.pie(labels2, values2, colors=colors2, title="Two Halves - Internal Fill Test")
    plt.show()

    print("\n" + "=" * 60)
    print("TWO SEGMENT ANALYSIS:")
    print("=" * 60)
    print("Each half should be COMPLETELY SOLID:")
    print("- Left half: solid blue with no black gaps inside")
    print("- Right half: solid orange with no black gaps inside")
    print("- Clean boundary between blue and orange")
    print("=" * 60)

    print("\nIf you see scattered individual blocks or black gaps")
    print("WITHIN the colored areas, the fill algorithm needs fixing.")

except ImportError as e:
    print(f"❌ Import error: {e}")
except Exception as e:
    print(f"❌ Error during testing: {e}")
    import traceback

    traceback.print_exc()



import plotext_plus._utility as ut

# this file builds a class inherited by the MonitorClass() in _monitor.py just because its only method - build_plot() - is very long and it is the core of plot building and it is written separately for clarity


class BuildClass:

    def build_plot(
        self,
    ):  # it builds the plot given the external settings and internal settings collected in draw()

        # Initial Tools
        r2 = [0, 1]
        signals = len(self.x)
        signal_indices = list(range(signals))
        texts = len(self.text)
        text_indices = list(range(texts))
        width, height = self.size
        ticks_colors = self.ticks_color, self.ticks_style

        # Find if Axes are used
        xside = [
            (self.default.xside[i] in self.xside + self.txside) or self.vcoord[i] != []
            for i in r2
        ]
        yside = [
            (self.default.yside[i] in self.yside + self.tyside) or self.hcoord[i] != []
            for i in r2
        ]

        # Remove Useless X and Y Ticks and Labels if axes are not used
        self.xticks = [self.xticks[i] if xside[i] else None for i in r2]
        self.xlabels = [self.xlabels[i] if xside[i] else None for i in r2]

        self.yticks = [self.yticks[i] if yside[i] else None for i in r2]
        self.ylabels = [self.ylabels[i] if yside[i] else None for i in r2]

        # Remove Useless h and v user defined Lines if axes are not used
        self.hcoord = [self.hcoord[i] if yside[i] else [] for i in r2]
        self.vcoord = [self.vcoord[i] if xside[i] else [] for i in r2]
        self.hcolors = [self.hcolors[i] if yside[i] else [] for i in r2]
        self.vcolors = [self.vcolors[i] if xside[i] else [] for i in r2]

        # Apply Scale (linear or log) to the data
        xscale = [
            ut.get_first(self.xscale, self.xside[s] is self.default.xside[0])
            for s in signal_indices
        ]  # the x scale for each signal
        yscale = [
            ut.get_first(self.yscale, self.yside[s] is self.default.yside[0])
            for s in signal_indices
        ]  # the y scale for each signal
        self.x = [
            ut.apply_scale(self.x[s], xscale[s] is self.default.xscale[1])
            for s in signal_indices
        ]  # apply optional log scale
        self.y = [
            ut.apply_scale(self.y[s], yscale[s] is self.default.yscale[1])
            for s in signal_indices
        ]

        # Apply Scale (linear or log) to the Axes Ticks
        self.xticks = [
            (
                ut.apply_scale(self.xticks[i], self.xscale[i] is self.default.xscale[1])
                if self.xticks[i] is not None
                else None
            )
            for i in r2
        ]  # apply optional log scale
        self.yticks = [
            (
                ut.apply_scale(self.yticks[i], self.yscale[i] is self.default.yscale[1])
                if self.yticks[i] is not None
                else None
            )
            for i in r2
        ]  # apply optional log scale

        # Apply Scale (linear or log) to the user defined Lines
        self.hcoord = [
            ut.apply_scale(self.hcoord[i], self.yscale[i] is self.default.yscale[1])
            for i in r2
        ]  # apply optional log scale
        self.vcoord = [
            ut.apply_scale(self.vcoord[i], self.xscale[i] is self.default.xscale[1])
            for i in r2
        ]  # apply optional log scale

        # Apply Scale (linear or log) to the user defined Text
        txscale = [
            ut.get_first(self.xscale, self.txside[s] is self.default.xside[0])
            for s in text_indices
        ]  # the x scale for each text
        tyscale = [
            ut.get_first(self.yscale, self.tyside[s] is self.default.yside[0])
            for s in text_indices
        ]  # the y scale for each text
        self.tx = [
            ut.apply_scale([self.tx[s]], txscale[s] is self.default.xscale[1])[0]
            for s in text_indices
        ]  # if width_canvas * height_canvas > 0 else [] # apply optional log scale
        self.ty = [
            ut.apply_scale([self.ty[s]], tyscale[s] is self.default.yscale[1])[0]
            for s in text_indices
        ]  # if width_canvas * height_canvas > 0 else [] # apply optional log scale
        tx = [
            [self.tx[s] for s in text_indices if self.txside[s] is self.default.xside[i]]
            for i in r2
        ]  # text x coord for each axis
        ty = [
            [self.ty[s] for s in text_indices if self.tyside[s] is self.default.yside[i]]
            for i in r2
        ]  # text x coofor each axis

        # Get X Axes Limits
        x = [
            ut.join([self.x[s] for s in signal_indices if self.xside[s] is side])
            for side in self.default.xside
        ]  # total x data for each axis
        x = [
            x[i] + self.vcoord[i] + tx[i] for i in r2
        ]  # add v lines and text coords to calculate xlim
        xlim = [ut.get_lim(el) if len(el) > 0 else [None, None] for el in x]
        self.xlim = [ut.replace_none(self.xlim[i], xlim[i]) for i in r2]
        self.xlim = [
            self.xlim[i][:: self.xdirection[i]] for i in r2
        ]  # optionally reverse axes
        xlim = [
            self.xlim[0] if self.xside[s] == self.default.xside[0] else self.xlim[1]
            for s in signal_indices
        ]  # xlim for each signal

        # Get Y Axes Limits
        y = [
            ut.join([self.y[s] for s in signal_indices if self.yside[s] is side])
            for side in self.default.yside
        ]
        y = [
            y[i] + self.hcoord[i] + ty[i] for i in r2
        ]  # add h lines and text coords to calculate ylim
        ylim = list(map(ut.get_lim, y))
        self.ylim = [ut.replace_none(self.ylim[i], ylim[i]) for i in r2]
        self.ylim = [
            self.ylim[i][:: self.ydirection[i]] for i in r2
        ]  # optionally reverse axes
        ylim = [
            self.ylim[0] if self.yside[s] == self.default.yside[0] else self.ylim[1]
            for s in signal_indices
        ]  # ylim for each signal

        # Get Y Ticks and Labels
        yticks_to_set = [
            self.yticks[i] is None and yside[i] and len(y[i]) > 0 for i in r2
        ]
        yticks = [
            (
                ut.linspace(*self.ylim[i], self.yfrequency[i])
                if yticks_to_set[i]
                else self.yticks[i]
            )
            for i in r2
        ]  # the actual Y ticks
        yticks_rescaled = [
            ut.reverse_scale(yticks[i], self.yscale[i] is self.default.yscale[1])
            for i in r2
        ]

        ylabels = [
            (
                self.date.times_to_string(yticks_rescaled[i])
                if self.y_date[i]
                else (
                    ut.get_labels(yticks_rescaled[i])
                    if yticks_to_set[i]
                    else self.ylabels[i]
                )
            )
            for i in r2
        ]
        ylabels = [
            (
                ut.add_extra_spaces(ylabels[i], self.default.yside[i])
                if ylabels[i] is not None
                else None
            )
            for i in r2
        ]
        width_ylabels = [ut.max_length(el) if el is not None else 0 for el in ylabels]

        # Get X Ticks and Labels
        xticks_to_set = [
            self.xticks[i] is None and xside[i] and len(x[i]) > 0 for i in r2
        ]
        xticks = [
            (
                ut.linspace(*self.xlim[i], self.xfrequency[i])
                if xticks_to_set[i]
                else self.xticks[i]
            )
            for i in r2
        ]  # the actual X ticks
        xticks_rescaled = [
            ut.reverse_scale(xticks[i], self.xscale[i] is self.default.xscale[1])
            for i in r2
        ]
        xlabels = [
            (
                self.date.times_to_string(xticks_rescaled[i])
                if self.x_date[i]
                else (
                    ut.get_labels(xticks_rescaled[i])
                    if xticks_to_set[i]
                    else self.xlabels[i]
                )
            )
            for i in r2
        ]
        xlabels = [
            (
                ut.add_extra_spaces(xlabels[i], self.default.xside[i])
                if xticks_to_set[i]
                else self.xlabels[i]
            )
            for i in r2
        ]
        height_xlabels = [len(el) > 0 if el is not None else 0 for el in xlabels]

        # Canvas Dimensions (the area of data points)
        width_canvas = width - sum(self.yaxes) - sum(width_ylabels)
        height_highbar = any(el is not None for el in [self.title, self.xlabel[1]])
        height_lowbar = any(el is not None for el in self.ylabel + [self.xlabel[0]])
        height_canvas = (
            height
            - sum(self.xaxes)
            - sum(height_xlabels)
            - height_highbar
            - height_lowbar
        )

        # Canvas Offset
        col_start = width_ylabels[0] + self.yaxes[0]
        col_end = col_start + width_canvas
        row_start = height_lowbar + height_xlabels[0] + self.xaxes[0]
        row_end = row_start + height_canvas

        # Get Absolute X and Y Ticks
        cticks = [
            (
                ut.get_matrix_data(xticks[i], self.xlim[i], width_canvas)
                if xticks[i] is not None
                else []
            )
            for i in r2
        ]
        rticks = [
            (
                ut.get_matrix_data(yticks[i], self.ylim[i], height_canvas)
                if yticks[i] is not None
                else []
            )
            for i in r2
        ]

        # Get Absolute Coordinates for user defined Lines
        hticks = [
            (
                ut.get_matrix_data(self.hcoord[i], self.ylim[i], height_canvas)
                if None not in self.ylim[i]
                else []
            )
            for i in r2
        ]
        vticks = [
            (
                ut.get_matrix_data(self.vcoord[i], self.xlim[i], width_canvas)
                if None not in self.xlim[i]
                else []
            )
            for i in r2
        ]

        # Get Absolute Coordinates for user defined Text
        txlim = [
            ut.get_first(self.xlim, self.txside[s] == self.default.xside[0])
            for s in text_indices
        ]  # xlim for each text
        tylim = [
            ut.get_first(self.ylim, self.tyside[s] == self.default.yside[0])
            for s in text_indices
        ]  # xlim for each text
        tcticks = [
            (
                ut.get_matrix_data([self.tx[s]], txlim[s], width_canvas)[0]
                if width_canvas > 0
                else 0
            )
            for s in text_indices
        ]  # if width_canvas > 0 else [] #
        trticks = [
            (
                ut.get_matrix_data([self.ty[s]], tylim[s], height_canvas)[0]
                if height_canvas > 0
                else 0
            )
            for s in text_indices
        ]  # if height_canvas > 0 else [] #

        # Get effective number of User Lines
        hlines = [len(el) for el in hticks]
        hline_indices = [
            list(range(el)) for el in hlines
        ]  # number of user defined horizontal lines for each y axis (as list)
        vlines = [len(el) for el in vticks]
        vline_indices = [
            list(range(el)) for el in vlines
        ]  # number of user defined vertical lines for each x axis (as list)

        # Create Matrix of Markers
        self.matrix.set_size(width, height)
        self.matrix.set_axes_color(self.axes_color)
        self.matrix.set_canvas_area(col_start, col_end, row_start, row_end)
        self.matrix.set_canvas_color(self.canvas_color)
        self.matrix.set_matrices()

        # Add Title
        col_center = (
            col_start + width_canvas // 2
        )  # centered on the canvas not the entire plot
        col_title = col_center if self.xlabel[1] is None else 0
        row_title = row_end + self.xaxes[1] + height_xlabels[1]
        alignment_title = "center" if self.xlabel[1] is None else "left"
        (
            self.matrix.add_horizontal_string(
                col_title,
                row_title,
                self.title,
                *ticks_colors,
                alignment=alignment_title,
                check_space=True,
            )
            if self.title and height > 0
            else None
        )

        # Add Upper X Label
        (
            self.matrix.add_horizontal_string(
                col_center,
                row_title,
                self.xlabel[1],
                *ticks_colors,
                alignment="center",
                check_space=True,
            )
            if self.xlabel[1] and height > 0
            else None
        )

        # Add Lower X Ticks
        row_xticks = min(row_start - self.xaxes[0] - 1, height - 1)
        cticks_inserted = [
            height > 0
            and self.matrix.add_horizontal_string(
                col_start + cticks[0][i],
                row_xticks,
                xlabels[0][i],
                *ticks_colors,
                alignment="dynamic",
                check_space=True,
            )
            for i in range(len(cticks[0]))
        ]
        cticks[0] = [
            cticks[0][i] for i in range(len(cticks[0])) if cticks_inserted[i]
        ]  # it updates the x ticks coordinates whatever x labels were inserted

        # Add Upper X Ticks: the reason to do it here prematurely, is that the cticks[1] need to be re-evaluated for next step
        row_xticks = row_end + self.xaxes[1]
        cticks_inserted = [
            height > 0
            and self.matrix.add_horizontal_string(
                col_start + cticks[1][i],
                row_xticks,
                xlabels[1][i],
                *ticks_colors,
                alignment="dynamic",
                check_space=True,
            )
            for i in range(len(cticks[1]))
        ]
        cticks[1] = [
            cticks[1][i] for i in range(len(cticks[1])) if cticks_inserted[i]
        ]  # it updates the x ticks coordinates whatever x labels were inserted

        # Add Upper X Axes (from previous step)
        def get_upper_x_tick(i):
            if self.grid[1] and i in cticks[1]:
                return "┼"
            elif self.grid[1] and i in cticks[0] or i in ut.join(vticks):
                return "┬"
            elif i in cticks[1]:
                return "┴"
            else:
                return "─"

        xaxis = [get_upper_x_tick(i) for i in range(width_canvas)]
        (
            self.matrix.add_horizontal_string(
                col_start, row_end, xaxis, self.ticks_color
            )
            if self.xaxes[0] and height_canvas >= -1
            else None
        )

        # Add Left Y Ticks
        (
            [
                self.matrix.add_horizontal_string(
                    0, rticks[0][i] + row_start, ylabels[0][i], *ticks_colors
                )
                for i in range(len(rticks[0]))
            ]
            if width >= width_ylabels[0]
            else None
        )

        # Add Left Y Axis
        def get_left_y_tick(i):
            if self.grid[0] and i in rticks[0]:
                return "┼"
            elif self.grid[0] and i in rticks[1] or i in ut.join(hticks):
                return "├"
            elif i in rticks[0]:
                return "┤"
            else:
                return "│"

        yaxis = [get_left_y_tick(i) for i in range(height_canvas)]
        col_yaxis = width_ylabels[0]
        (
            self.matrix.add_vertical_string(
                col_yaxis, row_start, yaxis, self.ticks_color
            )
            if self.yaxes[0] and width >= sum(width_ylabels) + 1
            else None
        )

        # Add Right Y Axis
        def tick(i):
            return (
                    "┼"
                    if (self.grid[0] and i in rticks[1])
                    else (
                        "┤"
                        if (self.grid[0] and i in rticks[0] or i in ut.join(hticks))
                        else "├" if i in rticks[1] else "│"
                    )
                )
        yaxis = [tick(i) for i in range(height_canvas)]
        (
            self.matrix.add_vertical_string(col_end, row_start, yaxis, self.ticks_color)
            if self.yaxes[1] and width >= sum(width_ylabels) + self.yaxes[0] + 1
            else None
        )

        # Add Right Y Ticks
        (
            [
                self.matrix.add_horizontal_string(
                    col_end + 1, rticks[1][i] + row_start, ylabels[1][i], *ticks_colors
                )
                for i in range(len(rticks[1]))
            ]
            if width >= sum(width_ylabels) + 1
            else None
        )

        # Add Frame 4 Corners if necessary
        canvas_test = height_canvas >= 0 and width_canvas >= 0
        (
            self.matrix.insert_element(
                col_start - 1, row_start - 1, "└", self.ticks_color
            )
            if self.xaxes[0] and self.yaxes[0] and canvas_test
            else None
        )
        (
            self.matrix.insert_element(col_end, row_start - 1, "┘", self.ticks_color)
            if self.xaxes[0] and self.yaxes[1] and canvas_test
            else None
        )
        (
            self.matrix.insert_element(col_start - 1, row_end, "┌", self.ticks_color)
            if self.xaxes[1] and self.yaxes[0] and canvas_test
            else None
        )
        (
            self.matrix.insert_element(col_end, row_end, "┐", self.ticks_color)
            if self.xaxes[1] and self.yaxes[1] and canvas_test
            else None
        )

        # Add Lower X Axes (from previous step)
        def tick(i):
            return (
                    "┼"
                    if (self.grid[1] and i in cticks[0])
                    else (
                        "┴"
                        if (self.grid[1] and i in cticks[1] or i in ut.join(vticks))
                        else "┬" if i in cticks[0] else "─"
                    )
                )
        xaxis = [tick(i) for i in range(width_canvas)]
        (
            self.matrix.add_horizontal_string(
                col_start, row_start - 1, xaxis, self.ticks_color
            )
            if self.xaxes[0] and height_canvas >= -1
            else None
        )

        # Add Left Y Label
        (
            self.matrix.add_horizontal_string(
                0, 0, self.ylabel[0], *ticks_colors, check_space=True
            )
            if self.ylabel[0] and height > 0
            else None
        )

        # Add Right Y Label
        (
            self.matrix.add_horizontal_string(
                width - 1,
                0,
                self.ylabel[1],
                *ticks_colors,
                check_space=True,
                alignment="right",
            )
            if self.ylabel[1] and height > 0
            else None
        )

        # Add Lower X Label
        (
            self.matrix.add_horizontal_string(
                col_center,
                0,
                self.xlabel[0],
                *ticks_colors,
                alignment="center",
                check_space=True,
            )
            if self.xlabel[0] and height > 0
            else None
        )

        # Add Grid Lines
        hline = "─" * width_canvas
        [
            self.matrix.add_horizontal_string(
                0 + col_start, row + row_start, hline, self.ticks_color
            )
            for row in ut.join(rticks)
            if self.grid[0]
        ]
        vline = "│" * height_canvas
        [
            self.matrix.add_vertical_string(
                col + col_start, 0 + row_start, vline, self.ticks_color
            )
            for col in ut.join(cticks)
            if self.grid[1]
        ]
        [
            self.matrix.insert_element(col + col_start, row + row_start, "┼")
            for row in ut.join(rticks)
            for col in ut.join(cticks)
            if all(self.grid)
        ]  # deals with the crossing between grids

        # Add user defined Lines
        [
            [
                self.matrix.add_horizontal_string(
                    col_start, hticks[i][line_idx] + row_start, hline, self.hcolors[i][line_idx]
                )
                for line_idx in hline_indices[i]
            ]
            for i in r2
        ]
        [
            [
                self.matrix.add_vertical_string(
                    vticks[i][line_idx] + col_start, 0 + row_start, vline, self.vcolors[i][line_idx]
                )
                for line_idx in vline_indices[i]
            ]
            for i in r2
        ]
        [
            [
                [
                    self.matrix.insert_element(
                        col + col_start,
                        hticks[i][line_idx] + row_start,
                        "┼",
                        self.hcolors[i][line_idx],
                    )
                    for line_idx in hline_indices[i]
                ]
                for i in r2
            ]
            for col in ut.join(cticks)
            if self.grid[1]
        ]  # deals with the crossing between h lines and v grids
        [
            [
                [
                    self.matrix.insert_element(
                        vticks[i][line_idx] + col_start,
                        row + row_start,
                        "┼",
                        self.vcolors[i][line_idx],
                    )
                    for line_idx in vline_indices[i]
                ]
                for i in r2
            ]
            for row in ut.join(rticks)
            if self.grid[0]
        ]  # deals with the crossing between v lines and h grids
        [
            [
                [
                    [
                        self.matrix.insert_element(
                            col_start + vticks[iv][lv],
                            hticks[i][line_idx] + row_start,
                            "┼",
                            self.hcolors[i][line_idx],
                        )
                        for line_idx in hline_indices[i]
                    ]
                    for i in r2
                ]
                for lv in vline_indices[iv]
            ]
            for iv in r2
        ]  # deals with the crossing between h and v lines

        # Expand Canvas to accommodate HD markers
        xf = [
            max([ut.marker_factor(el, 2, 2, 2) for el in self.marker[s]], default=1)
            for s in signal_indices
        ]
        yf = [
            max([ut.marker_factor(el, 2, 3, 4) for el in self.marker[s]], default=1)
            for s in signal_indices
        ]
        width_expanded = [width_canvas * el for el in xf]
        height_expanded = [height_canvas * el for el in yf]

        # Get Relative Data to Be Plotted on Matrix
        test_canvas = [width_expanded[s] * width_expanded[s] for s in signal_indices]
        x = [
            (
                ut.get_matrix_data(self.x[s], xlim[s], width_expanded[s])
                if test_canvas[s]
                else []
            )
            for s in signal_indices
        ]
        y = [
            (
                ut.get_matrix_data(self.y[s], ylim[s], height_expanded[s])
                if test_canvas[s]
                else []
            )
            for s in signal_indices
        ]
        m, c, st = self.marker, self.color, self.style

        # Add Lines between Data Points
        x, y, m, c, st = ut.transpose(
            [
                (
                    ut.get_lines(x[s], y[s], m[s], c[s], st[s])
                    if self.lines[s]
                    else (x[s], y[s], m[s], c[s], st[s])
                )
                for s in signal_indices
            ],
            5,
        )

        # Fillx
        # x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in signal_indices], 5)
        level = [
            ut.get_fill_level(self.fillx[s], ylim[s], height_expanded[s])
            for s in signal_indices
        ]
        x, y, m, c, st = ut.transpose(
            [
                (
                    ut.fill_data(x[s], y[s], level[s], m[s], c[s], st[s])
                    if self.fillx[s] is not False
                    else (x[s], y[s], m[s], c[s], st[s])
                )
                for s in signal_indices
            ],
            5,
        )

        # Filly
        # x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in signal_indices], 5)
        level = [
            ut.get_fill_level(self.filly[s], xlim[s], width_expanded[s])
            for s in signal_indices
        ]
        y, x, m, c, st = ut.transpose(
            [
                (
                    ut.fill_data(y[s], x[s], level[s], m[s], c[s], st[s])
                    if self.filly[s] is not False
                    else (y[s], x[s], m[s], c[s], st[s])
                )
                for s in signal_indices
            ],
            5,
        )

        # Get Actual HD Markers
        x, y, m, c, st = ut.transpose(
            [ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in signal_indices], 5
        )
        xf = [[ut.marker_factor(el, 2, 2, 2) for el in m[s]] for s in signal_indices]
        yf = [[ut.marker_factor(el, 2, 3, 4) for el in m[s]] for s in signal_indices]
        test = [max(xf[s], default=1) * max(yf[s], default=1) != 1 for s in signal_indices]
        x, y, mxy = ut.transpose(
            [
                ut.hd_group(x[s], y[s], xf[s], yf[s]) if test[s] else (x[s], y[s], [])
                for s in signal_indices
            ],
            3,
        )
        m = [
            [
                ut.get_hd_marker(mxy[s][i]) if m[s][i] in ut.hd_symbols else m[s][i]
                for i in range(len(x[s]))
            ]
            for s in signal_indices
        ]

        # Add Data to Canvas
        x, y, m, c, st = ut.transpose(
            [
                ut.remove_outsiders(
                    x[s], y[s], width_canvas, height_canvas, m[s], c[s], st[s]
                )
                for s in signal_indices
            ],
            5,
        )

        x, y, m, c, st = ut.transpose(
            [ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in signal_indices], 5
        )
        [
            [
                self.matrix.insert_element(
                    x[s][i] + col_start, y[s][i] + row_start, m[s][i], c[s][i], st[s][i]
                )
                for i in range(len(x[s]))
            ]
            for s in signal_indices
        ]

        # Legend Utilities
        def labelled(s):
            return self.label[s] is not None
        labels = [ut.space + self.label[s] + ut.space for s in signal_indices if labelled(s)]
        label_count = len(labels)
        max_label_length = ut.max_length(labels)
        labels = [el + ut.space * (max_label_length - len(el)) for el in labels]

        # Add Legend Side Symbols
        side = [
            ut.space + ut.side_symbols[(self.xside[s], self.yside[s])]
            for s in signal_indices
            if labelled(s)
        ]
        side_test = ut.no_duplicates(side) != [" L"] and label_count != 1
        side_width = 2 if side_test else 0
        legend_test = width_canvas >= side_width + 3 + max_label_length and height_canvas >= len(labels)
        (
            [
                self.matrix.add_horizontal_string(
                    col_start,
                    row_end - 1 - s,
                    side[s],
                    self.ticks_color,
                    self.ticks_style,
                )
                for s in range(label_count)
            ]
            if legend_test and side_test
            else None
        )

        # Add Legend Markers
        def take_3(data):
            return (data[:3] * 3)[:3]
        marker = [take_3(self.marker[s]) for s in signal_indices if labelled(s)]
        def replace_hd_marker(marker):
            return (
                    ut.hd_symbols.get(marker, marker)
                )
        marker = [[ut.space] + list(map(replace_hd_marker, el)) for el in marker]
        color = [[ut.no_color] + take_3(c[s]) for s in signal_indices if labelled(s)]
        style = [[ut.no_color] + take_3(st[s]) for s in signal_indices if labelled(s)]
        (
            [
                [
                    self.matrix.insert_element(
                        col_start + side_width + i,
                        row_end - 1 - s,
                        marker[s][i],
                        color[s][i],
                        style[s][i],
                    )
                    for i in range(3)
                ]
                for s in range(label_count)
            ]
            if legend_test
            else None
        )
        (
            [
                self.matrix.add_horizontal_string(
                    col_start + side_width + 3,
                    row_end - 1 - s,
                    labels[s],
                    self.ticks_color,
                    self.ticks_style,
                )
                for s in range(label_count)
            ]
            if legend_test
            else None
        )

        # Add Text to Canvas
        [
            self.matrix.add_multiple_horizontal_strings(
                col_start + tcticks[s],
                row_start + trticks[s],
                self.text[s],
                self.tfull[s],
                self.tstyle[s],
                self.tback[s],
                self.talign[s],
                False,
                True,
            )
            for s in text_indices
            if self.torien[s] is self.default.orientation[0]
        ]
        [
            self.matrix.add_multiple_vertical_strings(
                col_start + tcticks[s],
                row_start + trticks[s],
                self.text[s],
                self.tfull[s],
                self.tstyle[s],
                self.tback[s],
                self.talign[s],
                True,
            )
            for s in text_indices
            if self.torien[s] is self.default.orientation[1]
        ]



#!/usr/bin/env python3
"""
Basic plotting examples using the new Plotext API with chuk-term integration.
Demonstrates simple scatter plots, line plots, and bar charts.
"""

import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

import plotext_plus as plt


def basic_scatter_example():
    """Simple scatter plot example"""
    plt.log_info("Creating basic scatter plot...")

    # Sample data
    x = list(range(10))
    y = [i**2 for i in x]

    # Traditional API
    plt.scatter(x, y, color="blue")
    plt.title("Basic Scatter Plot - Traditional API")
    plt.xlabel("X Values")
    plt.ylabel("Y Values")
    plt.show()


def enhanced_scatter_example():
    """Enhanced scatter plot with banner mode"""
    plt.log_info("Creating enhanced scatter plot with banners...")

    # Sample data
    x = list(range(15))
    y = [i * 1.5 + (i % 3) * 2 for i in x]

    # New object-oriented API with banner
    (
        plt.Chart(use_banners=True, banner_title="📊 Enhanced Visualization")
        .scatter(x, y, color="red", label="Data Points")
        .title("Enhanced Scatter Plot")
        .xlabel("Time Steps")
        .ylabel("Measurements")
        .show()
    )


def line_plot_example():
    """Simple line plot example"""
    plt.log_info("Creating line plot...")

    # Sample data
    x = list(range(20))
    y = [i + (i % 5) * 1.5 for i in x]

    # Quick function API
    plt.quick_line(
        x,
        y,
        title="Quick Line Plot",
        xlabel="Time",
        ylabel="Value",
        use_banners=True,
        banner_title="📈 Trend Analysis",
    )


def bar_chart_example():
    """Simple bar chart example"""
    plt.log_info("Creating bar chart...")

    # Sample data
    categories = ["A", "B", "C", "D", "E"]
    values = [23, 17, 35, 29, 12]

    # Quick bar chart
    plt.quick_bar(
        categories,
        values,
        title="Sample Bar Chart",
        use_banners=True,
        banner_title="📊 Category Analysis",
    )


def multiple_series_example():
    """Multiple data series on one chart"""
    plt.log_info("Creating multi-series plot...")

    x = list(range(12))
    y1 = [i * 2 for i in x]
    y2 = [i**1.5 for i in x]

    # Object-oriented API with multiple series
    (
        plt.Chart(use_banners=True, banner_title="📈 Comparison Chart")
        .line(x, y1, color="blue", label="Linear")
        .scatter(x, y2, color="red", label="Power")
        .title("Multiple Data Series")
        .xlabel("X Axis")
        .ylabel("Y Axis")
        .show()
    )


def main():
    """Run all basic examples"""
    plt.log_success("🚀 Starting Basic Plotext Examples")

    basic_scatter_example()
    print("\n" + "=" * 50 + "\n")

    enhanced_scatter_example()
    print("\n" + "=" * 50 + "\n")

    line_plot_example()
    print("\n" + "=" * 50 + "\n")

    bar_chart_example()
    print("\n" + "=" * 50 + "\n")

    multiple_series_example()

    plt.log_success("✅ All basic examples completed!")


if __name__ == "__main__":
    main()
